# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.parser.utils
    :platform: Unix
    :synopsis: Burp-UI configuration file parser utilities.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import re
import codecs
import shutil

from copy import copy
from hashlib import md5
from collections import OrderedDict
from werkzeug.datastructures import MultiDict
from glob import glob
from six import iteritems, viewkeys

from ...utils import NOTIF_ERROR, NOTIF_OK, NOTIF_WARN, sanitize_string


class Option(object):
    """Object representing an option

    :param name: Option name
    :type name: str

    :param value: Option value
    :type value: str
    """
    type = None
    delim = '='
    _dirty = False

    def __init__(self, name, value=None):
        self.name = name
        self.value = value
        self._dirty = True

    @property
    def dirty(self):
        return self._dirty

    def update(self, value):
        """Change the option value"""
        self._dirty = True
        self.value = value

    def clean(self):
        """Mark the option as clean"""
        self._dirty = False

    def parse(self):
        """Parse the option value"""
        return self.value

    def dump(self):
        """Return the option representation to store in configuration file"""
        return "{} {} {}".format(self.name, self.delim, self.value)

    def __repr__(self):
        """Option representation"""
        return "{} -> {}".format(self.name, self.parse())

    def __str__(self):
        """Option to string"""
        return self.dump()


class OptionStr(Option):
    """Option type String

    Example:
        server = toto
    """
    type = 'string'


class OptionInt(Option):
    """Option type Integer

    Example:
        port = 1234
    """
    type = 'integer'

    def parse(self):
        """Parse the option value"""
        return int(self.value)


class OptionBool(Option):
    """Option type Boolean

    Example:
        hardlinked_archive = 1
    """
    type = 'boolean'

    def __init__(self, name, value=None):
        self.name = name
        self.value = self._format_value(value)
        self._dirty = True

    def update(self, value):
        """Change the option value"""
        self._dirty = True
        self.value = self._format_value(value)

    def _format_value(self, value):
        if self._parse(value):
            return 1
        return 0

    @staticmethod
    def _parse(value):
        try:
            if value is True:
                return True
            return int(value) == 1
        except ValueError:
            return False

    def parse(self):
        """Parse the option value"""
        return self._parse(self.value)


class OptionMulti(Option):
    """Option type Multi

    Example:
        keep = 7
        keep = 4
    """
    type = 'multi'

    def __init__(self, name, value=None):
        self.name = name
        self._dirty = True
        if value:
            self.value = [value]
        else:
            self.value = []

    def append(self, value):
        self._dirty = True
        self.value.append(value)
        return self.value

    def remove(self, value):
        self._dirty = True
        self.value.remove(value)
        return self.value

    def index(self, value):
        return self.value.index(value)

    def dump(self):
        """Return the option representation to store in configuration file"""
        ret = u''
        for val in self.value:
            ret += '{} {} {}\n'.format(self.name, self.delim, val)
        return ret.rstrip('\n')


class OptionInc(Option):
    """Option type Include

    Example:
        . incexc/windows
    """
    type = 'include'
    delim = ""

    def __init__(self, parser, name, value=None, root=None, mode='srv'):
        """
        :param parser: Parser instance
        :type parser: :class:`burpui.misc.parser.burp1.Parser`
        """
        super(OptionInc, self).__init__(name, value)
        self.parser = parser
        self.mode = mode
        self.extended = []
        self._dirty = True
        self._glob = False
        self._glob_dir = None
        self._glob_mtime = -1
        if root:
            self.root = os.path.dirname(root)
        else:
            self.root = None

    @property
    def dirty(self):
        return self._dirty or self._glob_changed

    def _path_absolute(self, path):
        absolute = path
        if not path.startswith('/'):
            if self.root:
                absolute = os.path.join(self.root, path)
            elif self.mode == 'srv':
                absolute = os.path.join(self.parser.root, path)
            else:
                absolute = os.path.join(self.parser.clientconfdir, path)
        return absolute

    @property
    def _glob_changed(self):
        if not self._glob or not self._glob_dir:
            return False
        mtime = os.path.getmtime(self._glob_dir)
        return mtime != self._glob_mtime

    def extend(self):
        """Helper function for the parsing"""
        if not self._dirty and self.extended and not self._glob_changed:
            return self.extended
        paths = []
        root = self._path_absolute(self.value)
        for path in glob(root):
            if self.parser._is_secure_path(path) and os.path.isfile(path) and \
                    not path.endswith('~') and not path.endswith('.back'):
                paths.append(path)
        # more than one match mean we have a glob, for sure
        if len(paths) > 1 or (len(paths) == 1 and paths[0] != root):
            self._glob = True
            self._glob_dir = os.path.dirname(root)
            self._glob_mtime = os.path.getmtime(self._glob_dir)
        else:
            self._glob = False
        self.clean()
        self.extended = paths
        return paths

    def parse(self):
        """Parse the option value"""
        return self.extend()

    def dump(self):
        """Return the option representation to store in configuration file"""
        if self.extend() and not self.parser.backend.enforce:
            return '. {}'.format(self.name)
        # if the include did not match anything, we can safely remove it
        return ''


class File(dict):
    """Object representing a configuration file

    :param parser: Parser object
    :type parser: :class:`burpui.misc.parser.doc.Doc`
    """
    md5 = None
    mtime = 0

    def __init__(self, parser, name=None, mode='srv', parent=None):
        """
        :param parser: Parser object
        :type parser: :class:`burpui.misc.parser.doc.Doc`

        :param name: File name
        :type name: str

        :param mode: Configuration type
        :type mode: str
        """
        # _dirty is used to know if the object changed
        self._dirty = False
        # _changed is used to know if the file changed since last read
        self._changed = True
        # cache the content of the file
        self._raw = []
        self._raw_data = MultiDict()
        self._data = MultiDict()
        self.parser = parser
        self.mode = mode
        self.name = name
        self.parent = parent
        self.options = OrderedDict()
        self.types = {
            'boolean': OrderedDict(),
            'integer': OrderedDict(),
            'include': OrderedDict(),
            'multi': OrderedDict(),
            'string': OrderedDict(),
        }
        if self.name:
            self.parse()

    @property
    def changed(self):
        for key, val in iteritems(self.types['include']):
            if val.dirty:
                self._changed = True
                return self._changed
        try:
            if self.name:
                mtime = os.path.getmtime(self.name)
            else:
                self._changed = True
                return True
        except os.error:
            self._changed = True
            return True
        if mtime != self.mtime:
            self._changed = self._md5 != self.md5
            return self._changed
        self._changed = mtime != self.mtime
        return self._changed

    def _ret_data(self, raw=True):
        if raw:
            ret = self._raw_data
        else:
            ret = self._data
        if ret and not self.dirty:
            return ret
        ret.clear()
        for key, val in iteritems(self.options):
            if isinstance(val, OptionMulti) and not raw:
                ret.setlist(key, val.parse())
            else:
                ret[key] = val if raw else val.parse()
        return ret

    @property
    def raw_data(self):
        return self._ret_data()

    @property
    def data(self):
        return self._ret_data(False)

    def clone(self):
        cpy = File(self.parser, name=self.name, mode=self.mode)
        cpy.options = copy(self.options)
        cpy.types = copy(self.types)
        return cpy

    def clean(self):
        self._dirty = False
        for _, opt in iteritems(self.options):
            opt.clean()

    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = name
        self.parse()

    def flatten(self, typ, listed=True, parse=True):
        self._refresh_types()
        if listed:
            return [
                {
                    'name': key,
                    'value': val.parse() if parse else val
                }
                for key, val in iteritems(self.types[typ])
            ]
        ret = OrderedDict()
        if typ in self.types:
            for key, val in iteritems(self.types[typ]):
                ret[key] = val.parse() if parse else val
        return ret

    @property
    def dirty(self):
        if not self._dirty:
            self._dirty = any([x.dirty for _, x in iteritems(self.options)])
        return self._dirty

    @property
    def boolean(self):
        return self.flatten('boolean')

    @property
    def integer(self):
        return self.flatten('integer')

    @property
    def include(self):
        return self.flatten('include')

    @property
    def multi(self):
        return self.flatten('multi')

    @property
    def string(self):
        return self.flatten('string')

    def _refresh_types(self):
        if self._dirty:
            for key in viewkeys(self.types):
                self.types[key] = OrderedDict()

            for key, opt in iteritems(self.options):
                self.types[opt.type][key] = opt

        self._dirty = False

    def _options_for_type(self, typ):
        return getattr(self.parser, '{}_{}'.format(typ, self.mode), [])

    def _type_for_option(self, opt):
        if opt == u'.':
            return 'include'

        for typ in ['boolean', 'integer', 'multi', 'string']:
            if opt in self._options_for_type(typ):
                return typ
        return None

    def _new_opt(self, key, value=None, typ=None):
        typ = typ or self._type_for_option(key)

        if typ == 'boolean':
            return OptionBool(key, value)
        if typ == 'integer':
            return OptionInt(key, value)
        if typ == 'multi':
            opt = self.options.get(key, OptionMulti(key))
            opt.append(value)
            return opt
        if typ == 'include':
            key = value
            opt = OptionInc(
                self.parser,
                key,
                value,
                root=self.name,
                mode=self.mode
            )

        return OptionStr(key, value)

    @property
    def _md5(self):
        """Compute the md5sum of the file"""
        hash_md5 = md5()
        try:
            with open(self.name, "rb") as bfile:
                for chunk in iter(lambda: bfile.read(4096), b""):
                    hash_md5.update(chunk)
            return hash_md5.hexdigest()
        except IOError:
            return None

    def get(self, key, default=None):
        try:
            return self.options[key]
        except KeyError:
            return default

    def __getitem__(self, key):
        return self.options[key]

    def __setitem__(self, key, value):
        self._dirty = True
        if key in self._options_for_type('boolean'):
            opt = OptionBool(key, value)
        elif key in self._options_for_type('integer'):
            opt = OptionInt(key, value)
        elif key in self._options_for_type('multi'):
            opt = self.options.get(key, OptionMulti(key))
            opt.append(value)
        elif key == u'.':
            key = value
            opt = OptionInc(
                self.parser,
                key,
                value,
                root=self.name,
                mode=self.mode
            )
        else:
            opt = OptionStr(key, value)
        self.options[key] = opt
        self.types[opt.type][key] = opt

    def __repr__(self):
        self._refresh_types()
        ret = u''
        for key, opts in iteritems(self.types):
            ret += '{} =>\n'.format(key)
            for _, opt in iteritems(opts):
                ret += '\t' + repr(opt) + '\n'
        return ret.rstrip('\n')

    def __str__(self):
        ret = u''
        for _, val in iteritems(self.options):
            tmp = str(val)
            if tmp:
                ret += tmp + '\n'
        return ret.rstrip('\n')

    def __len__(self):
        return len(self.options)

    def __delitem__(self, key):
        self._dirty = True
        del self.options[key]

    def clear(self):
        self._dirty = True
        for key in self.types.keys():
            self.types[key].clear()
        return self.options.clear()

    def copy(self):
        return self.options.copy()

    def has_key(self, k):
        return k in self.options

    def update(self, *args, **kwargs):
        self._dirty = True
        return self.options.update(*args, **kwargs)

    def keys(self):
        return self.options.keys()

    def values(self):
        return self.options.values()

    def items(self):
        return self.options.items()

    def pop(self, *args):
        self._dirty = True
        return self.options.pop(*args)

    def __cmp__(self, dic):
        return cmp(self.options, dic)

    def __contains__(self, item):
        return item in self.options

    def __iter__(self):
        return iter(self.options)

    def __unicode__(self):
        return unicode(self.__repr__())

    @property
    def raw(self):
        if self._raw and not self._changed and not self.changed:
            return self._raw
        if not self.name:
            return self._raw
        try:
            with codecs.open(self.name, 'r', 'utf-8', errors='ignore') as fil:
                self._raw = [x.rstrip('\n') for x in fil.readlines()]
        except IOError:
            return self._raw

        self._changed = False
        self.mtime = os.path.getmtime(self.name)
        self.md5 = self._md5
        return self._raw

    def _write_key(self, fil, key, data):
        self._changed = True

        strict = 'regex' not in key
        if key in self.parser.boolean_srv or key in self.parser.boolean_cli:
            val = 0
            obj = data.get(key)
            if obj == 'true' or (isinstance(obj, bool) and obj):
                val = 1
            fil.write('{} = {}\n'.format(key, val))
        elif key == '.':
            val = sanitize_string(data)
            fil.write('. {}\n'.format(val))
        elif key in self.parser.multi_srv or key in self.parser.multi_cli:
            for val in [sanitize_string(x, strict) for x in data.getlist(key)]:
                fil.write('{} = {}\n'.format(key, val))
        else:
            val = sanitize_string(str(data.get(key)), strict)
            fil.write('{} = {}\n'.format(key, val))

    def parse(self, force=False):
        """Parse the current config"""
        if not self._changed and not self.changed and not force:
            return

        self.clear()
        for line in self.raw:
            if re.match(r'^\s*#', line):
                continue
            res = re.search(r'\s*([^=\s]+)\s*=?\s*(.*)$', line)
            if res:
                key = res.group(1)
                val = res.group(2)
                if key == u'compression':
                    val = val.replace('zlib', 'gzip')
                elif key == u'ssl_compression':
                    val = val.replace('gzip', 'zlib')
                self[key] = val

        self._dirty = False

    def _store(self, data, dest=None, insecure=False):
        """Store the config"""
        dest = dest or self.name
        if not dest:
            return [[NOTIF_ERROR, 'No file defined!']]

        dirname = os.path.dirname(dest)
        filename = os.path.basename(dest)
        if dirname and not os.path.exists(dirname):
            try:
                os.makedirs(dirname)
            except OSError as exp:
                return [[NOTIF_WARN, str(exp)]]

        if not insecure:
            ref = os.path.join(dirname, '.{}.bui.init.back~'.format(filename))
            bak = os.path.join(dirname, '.{}.back~'.format(filename))

            if not os.path.isfile(ref) and os.path.isfile(dest):
                try:
                    shutil.copy(dest, ref)
                except IOError as exp:
                    return [[NOTIF_ERROR, str(exp)]]
            elif os.path.isfile(dest):
                try:
                    shutil.copy(dest, bak)
                except IOError as exp:
                    return [[NOTIF_ERROR, str(exp)]]

        errs = []

        for key in data.keys():
            if key in self.parser.files:
                dat = data.get(key)
                if not os.path.isfile(dat):
                    typ = 'strings'
                    if key in getattr(self.parser, 'multi_{}'.format(self.mode)):
                        typ = 'multis'
                    elif key in getattr(self.parser, 'boolean_{}'.format(self.mode)):
                        typ = 'bools'
                    elif key in getattr(self.parser, 'integer_{}'.format(self.mode)):
                        typ = 'integers'
                    # highlight the wrong parameters
                    errs.append([
                        NOTIF_ERROR,
                        "Sorry, the file '{}' does not exist".format(dat),
                        key,
                        typ
                    ])

        if errs and not insecure:
            return errs

        orig = self.raw
        oldkeys = [self._get_line_key(x) for x in orig]
        newkeys = list(set(viewkeys(data)) - set(oldkeys))

        already_multi = []
        already_file = []
        written = []
        skip = []

        # def lookup_option(key, val, strict=True):
        #    """returns a list of tuples (idx, line)"""
        #    if strict:
        #        reg = r'^\s*{}\s*=?\s*{}$'.format(key, val)
        #    else:
        #        reg = r'^\s*{}\s*=?'.format(key)
        #    ret = []
        #    for idx, line in enumerate(orig):
        #        if idx in skip:
        #            continue
        #        if re.match(reg, line):
        #            return idx, line
        #    return -1, None

        try:
            with codecs.open(dest, 'w', 'utf-8', errors='ignore') as fil:
                # f.write('# Auto-generated configuration using Burp-UI\n')
                for idx, line in enumerate(orig):
                    if idx in skip:
                        continue
                    if (self._line_removed(line, data.keys()) and
                            not self._line_is_comment(line) and
                            not self._line_is_file_include(line)):
                        # The line was removed, we comment it
                        fil.write('#{}\n'.format(line))
                    elif self._line_is_file_include(line):
                        # The line is a file inclusion, we check if the line
                        # was already present
                        ori = self._include_get_file(line)
                        if ori in data.getlist('includes_ori'):
                            idx = data.getlist('includes_ori').index(ori)
                            inc = data.getlist('includes')[idx]
                            self._write_key(fil, '.', inc)
                            already_file.append(inc)
                        else:
                            lead = ''
                            if not self._line_is_comment(line):
                                lead = '#'
                            fil.write('{}{}\n'.format(lead, line))

                    elif self._get_line_key(line, False) in data.keys():
                        # The line is still present or has been un-commented,
                        # rewrite it with eventual changes
                        key = self._get_line_key(line, False)
                        if key in written:
                            fil.write('#{}\n'.format(line))
                        else:
                            if key not in already_multi:
                                self._write_key(
                                    fil,
                                    key,
                                    data,
                                )
                            if key in getattr(self.parser, 'multi_{}'.format(self.mode)):
                                already_multi.append(key)
                            else:
                                written.append(key)
                    else:
                        lead = ''
                        if self._get_line_key(line, False) in written:
                            lead = '#'
                        fil.write('{}{}\n'.format(lead, line))
                # Write the new keys
                for key in newkeys:
                    if (key not in written and key not in already_multi and
                            key not in ['includes', 'includes_ori']):
                        self._write_key(
                            fil,
                            key,
                            data,
                        )
                # Write the rest of file inclusions
                if 'includes' in data:
                    for inc in data.getlist('includes'):
                        if inc not in already_file:
                            self._write_key(fil, '.', inc)
        except Exception as exp:
            return [[NOTIF_ERROR, str(exp)]]

        self.parse(True)

        return [[NOTIF_OK, 'Configuration successfully saved.']]

    def store(self, dest=None, insecure=False):
        """Store the current conf object"""
        data = self.data
        return self._store(data, dest, insecure)

    def store_data(self, data, insecure=False):
        """Store the conf given a object data"""
        return self._store(data, insecure=insecure)

    @staticmethod
    def _line_is_comment(line):
        """Check whether a given line is a comment or not"""
        if not line:
            return False
        return line.startswith('#')

    @staticmethod
    def _line_is_file_include(line):
        """Check whether a given line is a file inclusion or not"""
        if not line:
            return False
        return line.startswith('.') or re.match(r'^#+\s*\.', line) is not None

    @staticmethod
    def _include_get_file(line):
        """Return the path of the included file(s)"""
        if not line:
            return None
        _, fil = re.split(r'\s+', line, 1)
        return fil

    @staticmethod
    def _get_line_key(line, ignore_comments=True):
        """Return the key of a given line"""
        if not line:
            return ''
        if '=' not in line:
            return line
        (key, _) = re.split(r'\s*=\s*', line, 1)
        if not ignore_comments:
            key = key.strip('# ')
        return key.strip()

    @staticmethod
    def _line_removed(line, keys):
        """Check whether a given line has been removed in the updated version"""
        if not line:
            return False
        (key, _) = re.split(r'\s+|=', line, 1)
        key = key.strip()
        return key not in keys


class Config(File):
    """Object representing a configuration

    A config is like a virtual file so we can reuse some methods

    :param parser: Parser object
    :type parser: :class:`burpui.misc.parser.doc.Doc`
    """

    def __init__(self, path=None, parser=None, mode='srv'):
        """
        :param parser: Parser object
        :type parser: :class:`burpui.misc.parser.doc.Doc`

        :param mode: Configuration type
        :type mode: str
        """
        # we need an OrderedDict since the order of the configuration matters
        self.files = OrderedDict()
        self._tree = []
        self.default = path
        self.name = path
        self._includes = []
        self._dirty = True
        if path:
            self.files[path] = File(parser, path, mode=mode)
        super(Config, self).__init__(parser, path, mode)

    @property
    def changed(self):
        for path, conf in iteritems(self.files):
            if conf.changed:
                return True
        return False

    def _parse(self):

        orig = self.files
        for root, conf in iteritems(orig):
            conf.parse()
            for key, val in iteritems(conf.flatten('include', False)):
                for path in val:
                    if not os.path.isabs(path):
                        path = os.path.join(os.path.dirname(root), path)
                    self.add_file(path, root)
                    self._includes.append(path)

        # recursively parse the conf
        if orig != self.files:
            self._parse()

    def parse(self, force=False):
        if not self.changed and not force:
            return

        del self._includes[:]
        self._parse()

        removed = []
        orig = self.files
        for path, conf in iteritems(orig):
            if conf.parent and (conf.name not in self._includes or
               conf.name in removed):
                removed.append(path)
                self.del_file(path)

    @property
    def tree(self):
        if not self.changed and not self.dirty and self._tree:
            return self._tree

        # make sure to refresh files list
        self.parse(True)

        def __new_node(name, parent=None):
            dirname = os.path.dirname(name)
            basename = os.path.basename(name)
            return {
                'name': basename,
                'full': name,
                'dir': dirname,
                'parent': parent,
                'children': []
            }

        self._tree[:]
        dflt = self.get_default(True)
        temp = {}

        # retrieve the offset of the default conf
        offset = 0
        for idx, (path, _) in enumerate(iteritems(self.files)):
            if path == dflt.name:
                offset = idx
                break

        for idx, (top, conf) in enumerate(iteritems(self.files)):
            if idx < offset:
                continue
            node = __new_node(conf.name)
            for key, val in iteritems(conf.flatten('include', False)):
                for path in val:
                    if not os.path.isabs(path):
                        path = os.path.join(os.path.dirname(top), path)
                    node['children'].append(__new_node(path, node['full']))
            temp[conf.name] = node
        self._tree = [x for _, x in iteritems(temp)]
        return self._tree

    def store(self, conf=None, dest=None, insecure=False):
        ret = []
        if conf and conf in self.files:
            return self.files[conf].store(dest, insecure)
        for name, conf in iteritems(self.files):
            ret += conf.store(dest, insecure)
        return ret

    def store_data(self, conf, data, insecure=False):
        return self.get_file(conf).store_data(data, insecure)

    def clone(self):
        cpy = Config(self.name, self.parser, self.mode)
        for path, parsed in iteritems(self.files):
            if path == self.name:
                continue
            cpy.add_file(path)
        return cpy

    def set_default(self, path):
        self.default = path
        self.name = path
        try:
            default = self.get_default(True)
            if not self.parser and default:
                self.parser = getattr(self.get_default(), 'parser')
            if not self.mode and default:
                self.mode = getattr(self.get_default(), 'mode')
        except ValueError:
            pass

    def get_default(self, exc=False):
        if self.default:
            return self.get_file(self.default)
        if exc:
            raise ValueError('No default configuration found')
        return File(self.parser, mode=self.mode)

    def add_file(self, path=None, parent=None):
        idx = path or self.default
        if idx not in self.files:
            self.files[idx] = File(self.parser, idx, self.mode, parent)
            self._dirty = True
        return self.files[idx]

    def get_file(self, path):
        ret = self.files.get(path, File(self.parser, mode=self.mode))
        ret.parse()
        return ret

    def del_file(self, path):
        self._dirty = True
        del self.files[path]

    def list_files(self):
        return self.files.keys()

    def _refresh(self):
        if self._dirty or \
                any([x.dirty
                     for _, x in iteritems(self.files)]):

            # cleanup "caches"
            self.options.clear()
            del self.options
            self.options = OrderedDict()
            for key in viewkeys(self.types):
                del self.types[key]
                self.types[key] = OrderedDict()

            # now update caches with new values
            for _, fil in iteritems(self.files):
                self.options.update(fil.options)
                # FIXME: find a way to cache efficiently
                # fil.clean()

            for key, val in iteritems(self.options):
                self.types[val.type][key] = val

        self._dirty = False

    def _get(self, key, default=None, raw=False):
        self._refresh()
        try:
            obj = self.options[key]
        except KeyError:
            if self.parser and key in self.parser.defaults:
                obj = self._new_opt(key, self.parser.defaults[key])
            else:
                return default
        return obj if raw else obj.parse()

    def get_raw(self, key, default=None):
        return self._get(key, default, True)

    def get(self, key, default=None):
        return self._get(key, default)

    def getlist(self, key):
        obj = self._get(key, raw=True)
        if not obj:
            return []
        if isinstance(obj, OptionMulti):
            return obj.parse()
        else:
            return [obj.parse()]

    def __getitem__(self, key):
        self._refresh()
        return self.options[key]

    def __setitem__(self, key, value):
        self.get_default(True)[key] = value
        self._dirty = True

    def __repr__(self):
        self._refresh()
        ret = u''
        for key, fil in iteritems(self.files):
            ret += '>' * 5 + key + '<' * 5 + '\n'
            ret += repr(fil) + '\n'
        return ret.rstrip('\n')

    def __str__(self):
        self._refresh()
        return super(Config, self).__str__()

    def __len__(self):
        self._refresh()
        return len(self.options)

    def __delitem__(self, key):
        del self.get_default(True)[key]
        self._dirty = True

    def clear(self):
        self._dirty = True
        return self.files.clear()

    def copy(self):
        return self.files.copy()

    def has_key(self, k):
        self._refresh()
        return k in self.options

    def update(self, *args, **kwargs):
        self._dirty = True
        return self.get_default(True).update(*args, **kwargs)

    def keys(self):
        self._refresh()
        return self.options.keys()

    def values(self):
        self._refresh()
        return self.options.values()

    def items(self):
        self._refresh()
        return self.options.items()

    def pop(self, *args):
        self._dirty = True
        return self.get_default(True).pop(*args)

    def __cmp__(self, dic):
        self._refresh()
        return cmp(self.options, dic)

    def __contains__(self, item):
        self._refresh()
        return item in self.options

    def __iter__(self):
        self._refresh()
        return iter(self.options)

    def __unicode__(self):
        return unicode(repr(self))



# -*- coding: utf8 -*-
"""
.. module:: burpui.models
    :platform: Unix
    :synopsis: Burp-UI DB models module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import datetime

from .ext.sql import db
from flask import current_app
from .server import BUIServer  # noqa

app = current_app  # type: BUIServer


class Task(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    uuid = db.Column(db.String, unique=True)
    task = db.Column(db.String)
    user = db.Column(db.String, nullable=True)
    timestamp = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    expire = db.Column(db.DateTime, nullable=True)

    def __init__(self, uuid, task, user=None, expire=None):
        self.task = task
        self.uuid = uuid
        self.user = user
        if expire is not None:
            self.expire = datetime.datetime.utcnow() + expire

    def __repr__(self):
        return '<Task {} ({})>'.format(self.task, self.uuid)


class Session(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    uuid = db.Column(db.String, unique=True)
    user = db.Column(db.String)
    ip = db.Column(db.String, nullable=True)
    ua = db.Column(db.String, nullable=True)
    timestamp = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    expire = db.Column(db.DateTime, nullable=True)
    permanent = db.Column(db.Boolean)
    api = db.Column(db.Boolean)

    def __init__(self, uuid, user, ip=None, ua=None,
                 permanent=False, api=False):
        self.uuid = uuid
        self.user = user
        self.ip = ip
        self.ua = ua
        self.permanent = permanent
        self.api = api
        if self.permanent:
            self.expire = datetime.datetime.utcnow() + \
                app.permanent_session_lifetime

    def refresh(self, ip=None):
        self.timestamp = datetime.datetime.utcnow()
        if ip:
            self.ip = ip
        if self.permanent:
            self.expire = datetime.datetime.utcnow() + \
                app.permanent_session_lifetime
        db.session.commit()

    def __repr__(self):
        return '<Session {} ({}, {}, {})>'.format(
            self.uuid,
            self.user,
            self.ip,
            self.ua
        )


def test_database():
    # This is probably not optimal on huge databases
    Task.query.first()
    Session.query.first()



# -*- coding: utf8 -*-
"""
.. module:: burpui.forms
    :platform: Unix
    :synopsis: Burp-UI forms module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from .ext.i18n import LANGUAGES, get_locale

from flask_wtf import Form
from flask_babel import lazy_gettext as __
from wtforms import TextField, PasswordField, BooleanField, SelectField, validators


class LoginForm(Form):
    username = TextField(__('Username'), [validators.Length(min=2, max=25)])
    password = PasswordField(__('Password'), [validators.Required()])
    language = SelectField(__('Language'), choices=LANGUAGES.items(), default=get_locale)
    remember = BooleanField(__('Remember me'), [validators.Optional()])



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.async
    :platform: Unix
    :synopsis: Burp-UI async api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import select
import struct

from . import api, cache_key
from .misc import History
from .custom import Resource
from .client import ClientTreeAll, node_fields
from .clients import RunningBackup, ClientsReport
from ..exceptions import BUIserverException
from ..server import BUIServer  # noqa
from ..sessions import session_manager
from ..ext.async import celery
from ..ext.cache import cache
from ..config import config

from six import iteritems
from zlib import adler32
from flask import url_for, Response, current_app, after_this_request, \
    send_file, redirect
from time import gmtime, strftime, time, sleep
from datetime import timedelta, datetime
from werkzeug.datastructures import Headers
from celery.schedules import crontab
from celery.utils.log import get_task_logger

if config.get('WITH_SQL'):
    from ..ext.sql import db
else:
    db = None

bui = current_app  # type: BUIServer
ns = api.namespace('async', 'Asynchronous methods')
logger = get_task_logger(__name__)
ME = __name__

LOCK_EXPIRE = 60 * 30  # Lock expires in 30 minutes

BEAT_SCHEDULE = {
    'ping-backend-hourly': {
        'task': '{}.ping_backend'.format(ME),
        'schedule': crontab(minute='15'),  # run every hour
    },
    'backup-running-minutely': {
        'task': '{}.backup_running'.format(ME),
        'schedule': crontab(),  # run every minute
    },
    'get-all-backups-every-twenty-minutes': {
        'task': '{}.get_all_backups'.format(ME),
        'schedule': crontab(minute='*/20'),  # every 20 minutes
    },
    'get-all-clients-reports-every-twenty-minutes': {
        'task': '{}.get_all_clients_reports'.format(ME),
        'schedule': crontab(minute='*/20'),  # every 20 minutes
    },
    'cleanup-expired-sessions-daily': {
        'task': '{}.cleanup_expired_sessions'.format(ME),
        'schedule': crontab(hour='1'),  # every day at 1
    },
}

if db:
    from ..models import Task

    BEAT_SCHEDULE.update({
        'cleanup-restore-hourly': {
            'task': '{}.cleanup_restore'.format(ME),
            'schedule': crontab(minute='12'),  # run every hour
        },
    })

if 'CELERYBEAT_SCHEDULE' in celery.conf and \
        isinstance(celery.conf['CELERYBEAT_SCHEDULE'], dict):
    celery.conf['CELERYBEAT_SCHEDULE'].update(BEAT_SCHEDULE)
else:
    celery.conf['CELERYBEAT_SCHEDULE'] = BEAT_SCHEDULE


def acquire_lock(name, value='nyan', timeout=LOCK_EXPIRE):
    """Utility function to acquire a lock before processing the request"""
    lock = cache.cache.get(name)
    if lock:
        acquire_lock.lock = lock
        return False
    return cache.cache.add(name, value, timeout)


acquire_lock.lock = None


def release_lock(name):
    """Utility function to release a lock"""
    return cache.cache.delete(name)


def wait_for(lock_name, value, wait=10, timeout=LOCK_EXPIRE):
    """Utility function to wait until the given lock has been released"""
    old_lock = None
    if not acquire_lock(lock_name, value, timeout):
        logger.warn(
            'A task is already running. Wait for it: {}/{}'.format(
                lock_name,
                acquire_lock.lock
            )
        )
        old_lock = acquire_lock.lock
        # The lock should be released after LOCK_EXPLIRE max
        while not acquire_lock(lock_name, value, timeout):
            sleep(wait)

        # TODO: maybe we should check the status of task referenced by"old_lock"
        # to make sure the lock did not expire and the task is actually over
        logger.debug('lock released by: {}'.format(old_lock))

    return old_lock


@celery.task
def ping_backend():
    if bui.standalone:
        bui.client.status()
    else:
        def __status(server):
            (serv, back) = server
            try:
                return bui.client.status(agent=serv)
            except BUIserverException:
                return False

        map(
            __status,
            iteritems(bui.client.servers)
        )


@celery.task(bind=True)
def backup_running(self):
    # run once at the time, if one task was already running, we just discard
    # the new attempt
    if not acquire_lock(self.name):
        return None
    try:
        cache.cache.set(
            'backup_running_result',
            bui.client.is_one_backup_running(),
            120
        )
    finally:
        release_lock(self.name)


@celery.task(bind=True)
def get_all_backups(self):
    # run once at the time, if one task was already running, we just discard
    # the new attempt
    if not acquire_lock(self.name):
        return None
    try:
        backups = {}
        if bui.standalone:
            for cli in bui.client.get_all_clients():
                backups[cli['name']] = bui.client.get_client(cli['name'])
        else:
            for serv in bui.client.servers:
                backups[serv] = {}
                for cli in bui.client.get_all_clients(agent=serv):
                    backups[serv][cli['name']] = bui.client.get_client(cli['name'], agent=serv)
        cache.cache.set('all_backups', backups, 3600)
    finally:
        release_lock(self.name)


@celery.task(bind=True)
def get_all_clients_reports(self):
    # run once at the time, if one task was already running, we just discard
    # the new attempt
    if not acquire_lock(self.name):
        return None
    try:
        reports = {}
        if bui.standalone:
            reports = bui.client.get_clients_report(bui.client.get_all_clients())
        else:
            for serv in bui.client.servers:
                reports[serv] = bui.client.get_clients_report(bui.client.get_all_clients(agent=serv), serv)
        cache.cache.set('all_clients_reports', reports, 3600)
    finally:
        release_lock(self.name)


@celery.task
def cleanup_expired_sessions():
    def expires(sess):
        ret = session_manager.invalidate_session_by_id(sess.uuid)
        if ret:
            session_manager.delete_session_by_id(sess.uuid)
        return ret
    map(expires, session_manager.get_expired_sessions())


@celery.task
def cleanup_restore():
    tasks = db.session.query(Task).filter(Task.task == 'perform_restore').filter(datetime.utcnow() > Task.expire).all()
    # tasks = Task.query.filter_by(task='perform_restore').all()
    for rec in tasks:
        # if rec.expire and datetime.utcnow() > rec.expire:
        logger.info('Task expired: {}'.format(rec))
        task = perform_restore.AsyncResult(rec.uuid)
        try:
            if task.state != 'SUCCESS':
                logger.warn(
                    'Task is not done yet or did not end '
                    'successfully: {}'.format(task.state)
                )
                task.revoke(terminate=True)
                continue
            if not task.result:
                logger.warn('The task did not return anything')
                continue
            server = task.result.get('server')
            path = task.result.get('path')
            if path:
                if server:
                    if not bui.client.del_file(path, agent=server):
                        logger.warn("'{}' already removed".format(path))
                else:
                    if os.path.isfile(path):
                        os.unlink(path)
        finally:
            db.session.delete(rec)
            db.session.commit()
            task.revoke()


@celery.task(bind=True)
def perform_restore(self, client, backup,
                    files, strip, fmt, passwd, server=None, user=None,
                    expire=timedelta(minutes=60)):
    ret = None
    lock_name = '{}-{}'.format(self.name, server)

    # TODO: maybe do something with old_lock someday
    wait_for(lock_name, self.request.id)

    try:
        if server:
            filename = 'restoration_%d_%s_on_%s_at_%s.%s' % (
                backup,
                client,
                server,
                strftime("%Y-%m-%d_%H_%M_%S", gmtime()),
                fmt)
        else:
            filename = 'restoration_%d_%s_at_%s.%s' % (
                backup,
                client,
                strftime("%Y-%m-%d_%H_%M_%S", gmtime()),
                fmt)

        self.update_state(state='STARTED', meta={'step': 'doing'})
        archive, err = bui.client.restore_files(
            client,
            backup,
            files,
            strip,
            fmt,
            passwd,
            server
        )
        if not archive:
            if not err:
                err = 'Something went wrong while restoring'
            self.update_state(state='FAILURE', meta={'error': err})
            logger.error('FAILURE: {}'.format(err))
        else:
            ret = {
                'filename': filename,
                'path': archive,
                'user': user,
                'server': server
            }
            logger.debug(ret)

    finally:
        release_lock(lock_name)

    if db:
        curr = Task.query.filter_by(uuid=self.request.id).first()
        if curr:
            curr.expire = datetime.utcnow() + expire
            db.session.commit()

    if err:
        # make the task crash
        raise Exception(err)

    return ret


@celery.task(bind=True)
def load_all_tree(self, client, backup, server=None, user=None):
    key = 'load_all_tree-{}-{}-{}'.format(client, backup, server)
    ret = cache.cache.get(key)
    if ret:
        return {
            'client': client,
            'backup': backup,
            'server': server,
            'user': user,
            'tree': ret
        }

    lock_name = '{}-{}'.format(self.name, server)

    # TODO: maybe do something with old_lock someday
    wait_for(lock_name, self.request.id)

    try:
        ret = ClientTreeAll._get_tree_all(client, backup, server)
    except BUIserverException as exp:
        raise Exception(str(exp))
    finally:
        release_lock(lock_name)

    cache.cache.set(key, ret, 3600)
    return {
        'client': client,
        'backup': backup,
        'server': server,
        'user': user,
        'tree': ret
    }


def force_scheduling_now():
    """Force scheduling some tasks now"""
    get_all_backups.delay()
    backup_running.delay()
    get_all_clients_reports.delay()


@ns.route('/status/<task_id>', endpoint='async_restore_status')
@ns.doc(
    params={
        'task_id': 'The task ID to process',
    }
)
class AsyncRestoreStatus(Resource):
    """The :class:`burpui.api.async.AsyncRestoreStatus` resource allows you to
    follow a restore task.

    This resource is part of the :mod:`burpui.api.async` module.
    """
    @ns.doc(
        responses={
            200: 'Success',
            500: 'Task failed',
        },
    )
    def get(self, task_id):
        """Returns the state of the given task"""
        task = perform_restore.AsyncResult(task_id)
        if task.state == 'FAILURE':
            if db:
                rec = Task.query.filter_by(uuid=task_id).first()
                if rec:
                    db.session.delete(rec)
                    db.session.commit()
            task.revoke()
            err = str(task.result)
            self.abort(502, err)
        if task.state == 'SUCCESS':
            if not task.result:
                self.abort(500, 'The task did not return anything')
            server = task.result.get('server')
            return {
                'state': task.state,
                'location': url_for(
                    '.async_get_file',
                    task_id=task_id,
                    server=server
                )
            }
        return {'state': task.state}


@ns.route('/get/<task_id>',
          '/<server>/get/<task_id>',
          endpoint='async_get_file')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'task_id': 'The task ID to process',
    }
)
class AsyncGetFile(Resource):
    """The :class:`burpui.api.async.AsyncGetFile` resource allows you to
    retrieve the archive generated by the given task.

    This resource is part of the :mod:`burpui.api.async` module.
    """

    @ns.doc(
        responses={
            200: 'Success',
            400: 'Incomplete task',
            403: 'Insufficient permissions',
            500: 'Task failed',
        },
    )
    def get(self, task_id, server=None):
        """Returns the generated archive"""
        task = perform_restore.AsyncResult(task_id)
        if task.state != 'SUCCESS':
            if task.state == 'FAILURE':
                self.abort(
                    500,
                    'Unsuccessful task: {}'.format(task.result.get('error'))
                )
            self.abort(400, 'Task not processed yet: {}'.format(task.state))

        path = task.result.get('path')
        user = task.result.get('user')
        dst_server = task.result.get('server')
        filename = task.result.get('filename')

        if self.username != user or (dst_server and dst_server != server):
            self.abort(403, 'Unauthorized access')

        if db:
            rec = Task.query.filter_by(uuid=task_id).first()
            if rec:
                db.session.delete(rec)
                db.session.commit()
        task.revoke()

        if dst_server:
            return self.stream_file(path, filename, dst_server)

        try:
            # Trick to delete the file while sending it to the client.
            # First, we open the file in reading mode so that a file handler
            # is open on the file. Then we delete it as soon as the request
            # ended. Because the fh is open, the file will be actually removed
            # when the transfer is done and the send_file method has closed
            # the fh. Only tested on Linux systems.
            fh = open(path, 'r')

            @after_this_request
            def remove_file(response):
                """Callback function to run after the client has handled
                the request to remove temporary files.
                """
                os.remove(path)
                return response

            resp = send_file(fh,
                             as_attachment=True,
                             attachment_filename=filename,
                             mimetype='application/zip')
            resp.set_cookie('fileDownload', 'true')
        except Exception as e:
            bui.client.logger.error(str(e))
            self.abort(500, str(e))

        return resp

    def stream_file(self, path, filename, server):
        socket = bui.client.get_file(path, server)
        if not socket:
            self.abort(500)
        lengthbuf = socket.recv(8)
        length, = struct.unpack('!Q', lengthbuf)

        def stream(sock, l):
            """The restoration took place on another server so we need
            to stream the file that is not present on the current
            machine.
            """
            bsize = 1024
            received = 0
            if l < bsize:
                bsize = l
            while received < l:
                buf = b''
                r, _, _ = select.select([sock], [], [], 5)
                if not r:
                    self.abort(500, 'Socket timed-out')
                buf += sock.recv(bsize)
                if not buf:
                    continue
                received += len(buf)
                self.logger.debug('{}/{}'.format(received, l))
                yield buf
            sock.sendall(struct.pack('!Q', 2))
            sock.sendall(b'RE')
            sock.close()

        headers = Headers()
        headers.add('Content-Disposition',
                    'attachement',
                    filename=filename)
        headers['Content-Length'] = length

        resp = Response(stream(socket, length),
                        mimetype='application/zip',
                        headers=headers,
                        direct_passthrough=True)
        resp.set_cookie('fileDownload', 'true')
        resp.set_etag('flask-%s-%s-%s' % (
            time(),
            length,
            adler32(filename.encode('utf-8')) & 0xffffffff))

        return resp


@ns.route('/archive/<name>/<int:backup>',
          '/<server>/archive/<name>/<int:backup>',
          endpoint='async_restore')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'name': 'Client name',
        'backup': 'Backup number',
    },
)
class AsyncRestore(Resource):
    """The :class:`burpui.api.async.AsyncRestore` resource allows you to
    perform a file restoration.

    This resource is part of the :mod:`burpui.api.async` module.

    The following parameters are supported:
    - ``list``: list of files/directories to restore
    - ``strip``: number of elements to strip in the path
    - ``format``: returning archive format
    - ``pass``: password to use for encrypted backups
    """
    parser = ns.parser()
    parser.add_argument(
        'pass',
        help='Password to use for encrypted backups',
        nullable=True
    )
    parser.add_argument(
        'format',
        required=False,
        help='Returning archive format',
        choices=('zip', 'tar.gz', 'tar.bz2'),
        default='zip',
        nullable=True
    )
    parser.add_argument(
        'strip',
        type=int,
        help='Number of elements to strip in the path',
        default=0,
        nullable=True
    )
    parser.add_argument(
        'list',
        required=True,
        help='List of files/directories to restore',
        nullable=False
    )

    @ns.expect(parser, validate=True)
    @ns.doc(
        responses={
            202: 'Accepted',
            400: 'Missing parameter',
            403: 'Insufficient permissions',
        },
    )
    def post(self, server=None, name=None, backup=None):
        """Performs an online restoration

        **POST** method provided by the webservice.

        This method returns a :mod:`flask.Response` object.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int
        """
        args = self.parser.parse_args()
        files = args['list']
        strip = args['strip']
        fmt = args['format'] or 'zip'
        passwd = args['pass']
        if not files or not name or not backup:
            self.abort(400, 'missing arguments')
        # Manage ACL
        if (bui.acl and
                (not bui.acl.is_client_allowed(self.username,
                                               name,
                                               server) and not
                 self.is_admin)):
            self.abort(
                403,
                'You are not allowed to perform a restoration for this client'
            )
        task = perform_restore.apply_async(
            args=[
                name,
                backup,
                files,
                strip,
                fmt,
                passwd,
                server,
                self.username
            ]
        )
        if db:
            db_task = Task(
                task.id,
                'perform_restore',
                self.username,
                timedelta(minutes=60)
            )
            db.session.add(db_task)
            db.session.commit()
        return {'id': task.id, 'name': 'perform_restore'}, 202


@ns.route('/backup-running',
          '/<server>/backup-running',
          endpoint='async_running_backup')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    }
)
class AsyncRunningBackup(RunningBackup):
    """The :class:`burpui.api.async.AsyncRunningBackup` resource allows you to
    access the status of the server in order to know if there is a running
    backup currently.

    This resource is backed by a periodic task. If the periodic task fail or is
    not running, we redirect to the "synchronous" API call.

    This resource is part of the :mod:`burpui.api.async` module.
    """

    @ns.marshal_with(
        RunningBackup.running_fields,
        code=200,
        description='Success',
        strict=False
    )
    def get(self, server=None):
        """Tells if a backup is running right now

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
                "running": false
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :returns: The *JSON* described above.
        """
        res = cache.cache.get('backup_running_result')
        if res is None:
            # redirect to synchronous API call
            # FIXME: Since we subclass the original code, we don't need the
            # redirect anymore if the redirection is problematic
            return redirect(url_for('api.running_backup', server=server))
        return {'running': self._is_one_backup_running(res, server)}


@ns.route('/history',
          '/history/<client>',
          '/<server>/history',
          '/<server>/history/<client>',
          endpoint='async_history')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'client': 'Client name',
    },
)
class AsyncHistory(History):
    """The :class:`burpui.api.misc.History` resource allows you to retrieve
    an history of the backups

    This resource is backed by a periodic task. If the periodic task fail or is
    not running, we redirect to the "synchronous" API call.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode and ``clientName`` is also allowed to filter
    by client.

    ::

        $('#calendar').fullCalendar({

            eventSources: [

                // your event source
                {
                    events: [ // put the array in the `events` property
                        {
                            title  : 'event1',
                            start  : '2010-01-01'
                        },
                        {
                            title  : 'event2',
                            start  : '2010-01-05',
                            end    : '2010-01-07'
                        },
                        {
                            title  : 'event3',
                            start  : '2010-01-09T12:30:00',
                        }
                    ],
                    color: 'black',     // an option!
                    textColor: 'yellow' // an option!
                }

                // any other event sources...

            ]

        });

    """

    @cache.cached(timeout=1800, key_prefix=cache_key)
    @ns.marshal_with(
        History.history_fields,
        code=200,
        description='Success',
        strict=False,
        as_list=True
    )
    @ns.expect(History.parser)
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
        },
    )
    def get(self, client=None, server=None):
        """Returns a list of calendars describing the backups that have been
        completed so far

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                "color": "#7C6F44",
                "events": [
                  {
                    "backup": "0000001",
                    "end": "2015-01-25 13:32:04+01:00",
                    "name": "toto-test",
                    "start": "2015-01-25 13:32:00+01:00",
                    "title": "Client: toto-test, Backup n°0000001",
                    "url": "/client/toto-test"
                  }
                ],
                "name": "toto-test",
                "textColor": "white"
              }
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str
        :param client: Which client to collect data from
        :type client: str

        :returns: The *JSON* described above
        """
        self._check_acl(client, server)
        res = cache.cache.get('all_backups')
        args = self.parser.parse_args()
        if res is None:
            # redirect to synchronous API call
            # FIXME: Since we subclass the original code, we don't need the
            # redirect anymore if the redirection is problematic
            return redirect(
                url_for(
                    'api.history',
                    client=client,
                    server=server,
                    start=args['start'],
                    end=args['end']
                )
            )

        return self._get_backup_history(client, server, res)


@ns.route('/report',
          '/<server>/report',
          endpoint='async_clients_report')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class AsyncClientsReport(ClientsReport):
    """The :class:`burpui.api.async.AsyncClientsReport` resource allows you to
    access general reports about your clients.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """

    @api.cache.cached(timeout=1800, key_prefix=cache_key)
    @ns.marshal_with(
        ClientsReport.report_fields,
        code=200,
        description='Success',
        strict=False
    )
    @ns.expect(ClientsReport.parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    def get(self, server=None):
        """Returns a global report about all the clients of a given server

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "backups": [
                {
                  "name": "client1",
                  "number": 15
                },
                {
                  "name": "client2",
                  "number": 1
                }
              ],
              "clients": [
                {
                  "name": "client1",
                  "stats": {
                    "total": 296377,
                    "totsize": 57055793698,
                    "windows": "unknown"
                  }
                },
                {
                  "name": "client2",
                  "stats": {
                    "total": 3117,
                    "totsize": 5345361,
                    "windows": "true"
                  }
                }
              ]
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :returns: The *JSON* described above
        """
        server = server or self.parser.parse_args()['serverName']
        self._check_acl(server)
        res = cache.cache.get('all_clients_reports')
        if res is None:
            # redirect to synchronous API call
            # FIXME: Since we subclass the original code, we don't need the
            # redirect anymore if the redirection is problematic
            return redirect(url_for('api.clients_report', server=server))
        return self._get_clients_reports(res, server)


@ns.route('/browseall/<name>/<int:backup>',
          '/<server>/browsall/<name>/<int:backup>',
          endpoint='async_client_tree_all')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in' +
                  ' multi-agent mode',
        'name': 'Client name',
        'backup': 'Backup number',
    },
)
class AsyncClientTreeAll(Resource):
    """The :class:`burpui.api.async.AsyncClientTreeAll` resource allows you to
    retrieve a list of all the files in a given backup through the celery
    worker.

    This resource is part of the :mod:`burpui.api.client` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """
    parser = ns.parser()
    parser.add_argument(
        'serverName',
        help='Which server to collect data from when in multi-agent mode'
    )

    @ns.expect(parser)
    @ns.doc(
        responses={
            202: 'Accepted',
            405: 'Method not allowed',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    def post(self, server=None, name=None, backup=None):
        """Launch the tasks that will gather all nodes of a given backup

        **POST** method provided by the webservice.

        This method returns a :mod:`flask.Response` object.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int
        """
        args = self.parser.parse_args()
        server = server or args.get('serverName')

        if not bui.client.get_attr('batch_list_supported', False, server):
            self.abort(
                405,
                'Sorry, the requested backend does not support this method'
            )

        # Manage ACL
        if (bui.acl and
                (not bui.acl.is_client_allowed(self.username,
                                               name,
                                               server) and not
                 self.is_admin)):
            self.abort(403, 'Sorry, you are not allowed to view this client')

        task = load_all_tree.apply_async(
            args=[
                name,
                backup,
                server,
                self.username
            ]
        )
        return {'id': task.id, 'name': 'load_all_tree'}, 202


@ns.route('/browse-status/<task_id>', endpoint='async_browse_status')
@ns.doc(
    params={
        'task_id': 'The task ID to process',
    }
)
class AsyncBrowseStatus(Resource):
    """The :class:`burpui.api.async.AsyncBrowseStatus` resource allows you to
    follow a browse task.

    This resource is part of the :mod:`burpui.api.async` module.
    """
    @ns.doc(
        responses={
            200: 'Success',
            500: 'Task failed',
        },
    )
    def get(self, task_id):
        """Returns the state of the given task"""
        task = load_all_tree.AsyncResult(task_id)
        if task.state == 'FAILURE':
            task.revoke()
            err = str(task.result)
            self.abort(502, err)
        if task.state == 'SUCCESS':
            if not task.result:
                self.abort(500, 'The task did not return anything')
            server = task.result.get('server')
            return {
                'state': task.state,
                'location': url_for(
                    '.async_do_browse_all',
                    task_id=task_id,
                    server=server
                )
            }
        return {'state': task.state}


@ns.route('/get-browse/<task_id>',
          '/<server>/get-browse/<task_id>',
          endpoint='async_do_browse_all')
@ns.doc(
    params={
        'task_id': 'The task ID to process',
    }
)
class AsyncDoBrowseAll(Resource):
    """The :class:`burpui.api.async.AsyncDoBrowseAll` resource allows you to
    retrieve the tree generated by the given task.

    This resource is part of the :mod:`burpui.api.async` module.
    """

    @ns.marshal_list_with(node_fields, code=200, description='Success')
    @ns.doc(
        responses={
            400: 'Incomplete task',
            403: 'Insufficient permissions',
            500: 'Task failed',
        },
    )
    def get(self, task_id, server=None):
        """Returns the generated archive"""
        task = load_all_tree.AsyncResult(task_id)
        if task.state != 'SUCCESS':
            if task.state == 'FAILURE':
                self.abort(
                    500,
                    'Unsuccessful task: {}'.format(task.result.get('error'))
                )
            self.abort(400, 'Task not processed yet: {}'.format(task.state))

        user = task.result.get('user')
        dst_server = task.result.get('server')
        resp = task.result.get('tree')

        if self.username != user or (dst_server and dst_server != server):
            self.abort(403, 'Unauthorized access')

        task.revoke()

        return resp



# -*- coding: utf8 -*-
"""
.. module:: burpui.utils
    :platform: Unix
    :synopsis: Burp-UI utils module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import math
import string
import sys
import datetime
import zipfile
import tarfile
import logging

from inspect import currentframe, getouterframes
from ._compat import PY3

NOTIF_OK = 0
NOTIF_WARN = 1
NOTIF_ERROR = 2
NOTIF_INFO = 3

if PY3:
    long = int  # pragma: no cover
    basestring = str  # pragma: no cover


class human_readable(long):
    """define a human_readable class to allow custom formatting
    format specifiers supported :
        em : formats the size as bits in IEC format i.e. 1024 bits (128 bytes) = 1Kib
        eM : formats the size as Bytes in IEC format i.e. 1024 bytes = 1KiB
        sm : formats the size as bits in SI format i.e. 1000 bits = 1kb
        sM : formats the size as bytes in SI format i.e. 1000 bytes = 1KB
        cm : format the size as bit in the common format i.e. 1024 bits (128 bytes) = 1Kb
        cM : format the size as bytes in the common format i.e. 1024 bytes = 1KB

    code from: http://code.activestate.com/recipes/578323-human-readable-filememory-sizes-v2/
    """
    def __format__(self, fmt):  # pragma: no cover
        # is it an empty format or not a special format for the size class
        if fmt == "" or fmt[-2:].lower() not in ["em", "sm", "cm"]:
            if fmt[-1].lower() in \
                    ['b', 'c', 'd', 'o', 'x', 'n', 'e', 'f', 'g', '%']:
                # Numeric format.
                return long(self).__format__(fmt)
            else:
                return str(self).__format__(fmt)

        if sys.version_info >= (3, 0):
            chars = string.ascii_lowercase
        else:
            chars = string.lowercase
        # work out the scale, suffix and base
        factor, suffix = (8, "b") if fmt[-1] in chars else (1, "B")
        base = 1024 if fmt[-2] in ["e", "c"] else 1000

        # Add the i for the IEC format
        suffix = "i" + suffix if fmt[-2] == "e" else suffix

        mult = ["", "K", "M", "G", "T", "P"]

        val = float(self) * factor
        i = 0 if val < 1 else int(math.log(val, base)) + 1
        v = val / math.pow(base, i)
        v, i = (v, i) if v > 0.5 else (v * base, i - 1)

        # Identify if there is a width and extract it
        width = "" if fmt.find(".") == -1 else fmt[:fmt.index(".")]
        precis = fmt[:-2] if width == "" else fmt[fmt.index("."):-2]

        # do the precision bit first, so width/alignment works with the suffix
        if float(self) == 0:
            return "{0:{1}f}".format(v, precis)
        t = ("{0:{1}f}" + mult[i] + suffix).format(v, precis)

        return "{0:{1}}".format(t, width) if width != "" else t


if PY3:  # pragma: no cover
    class BUIlogger(logging.Logger):
        padding = 0
        """Logger class for more convenience"""
        def makeRecord(self, name, level, fn, lno, msg,
                       args, exc_info, func=None, extra=None, sinfo=None):
            """
            Try to guess where was call the function
            """
            cf = currentframe()
            caller = getouterframes(cf)
            cpt = 0
            size = len(caller)
            me = __file__
            if me.endswith('.pyc'):
                me = me[:-1]
            # It's easy to get the _logger parent function because it's the
            # following frame
            while cpt < size - 1:
                (_, filename, _, function_name, _, _) = caller[cpt]
                if function_name == '_logger' and filename == me:
                    cpt += 1
                    break
                cpt += 1
            cpt += self.padding
            (frame, filename, line_number, function_name, lines, index) = \
                caller[cpt]
            return super(BUIlogger, self).makeRecord(
                name,
                level,
                filename,
                line_number,
                msg,
                args,
                exc_info,
                func=function_name,
                extra=extra,
                sinfo=sinfo
            )
else:
    class BUIlogger(logging.Logger):
        padding = 0
        """Logger class for more convenience"""
        def makeRecord(self, name, level, fn, lno, msg,
                       args, exc_info, func=None, extra=None):
            """Try to guess where was call the function"""
            cf = currentframe()
            caller = getouterframes(cf)
            cpt = 0
            size = len(caller)
            me = __file__
            if me.endswith('.pyc'):
                me = me[:-1]
            # It's easy to get the _logger parent function because it's the
            # following frame
            while cpt < size - 1:
                (_, filename, _, function_name, _, _) = caller[cpt]
                if function_name == '_logger' and filename == me:
                    cpt += 1
                    break
                cpt += 1
            cpt += self.padding
            (frame, filename, line_number, function_name, lines, index) = \
                caller[cpt]
            return super(BUIlogger, self).makeRecord(
                name,
                level,
                filename,
                line_number,
                msg,
                args,
                exc_info,
                func=function_name,
                extra=extra
            )


class BUIlogging(object):
    logger = None
    monkey = None
    padding = 0
    """Provides a generic logging method for all modules"""
    def _logger(self, level, msg, *args):
        """generic logging method so that the logging is backend-independent"""
        if (self.logger and
                self.logger.getEffectiveLevel() <= logging.getLevelName(
                    level.upper()
                )):
            sav = None
            if not self.monkey:
                self.monkey = BUIlogger(__name__)
            # bui-agent overrides the _logger function so we add a padding
            # offset
            self.monkey.padding = self.padding
            # dynamically monkey-patch the makeRecord function
            sav = self.logger.makeRecord
            self.logger.makeRecord = self.monkey.makeRecord
            self.logger.log(logging.getLevelName(level.upper()), msg, *args)
            self.logger.makeRecord = sav


class BUIcompress():
    """Provides a context to generate any kind of archive supported by
    burp-ui
    """
    def __init__(self, name, archive, zip64=False):  # pragma: no cover
        self.name = name
        self.archive = archive
        self.zip64 = zip64

    def __enter__(self):
        self.arch = None
        if self.archive == 'zip':
            self.arch = zipfile.ZipFile(
                self.name,
                mode='w',
                compression=zipfile.ZIP_DEFLATED,
                allowZip64=self.zip64
            )
        elif self.archive == 'tar.gz':
            self.arch = tarfile.open(self.name, 'w:gz')
        elif self.archive == 'tar.bz2':
            self.arch = tarfile.open(self.name, 'w:bz2')
        return self

    def __exit__(self, type, value, traceback):
        self.arch.close()

    def append(self, path, arcname):
        if self.archive == 'zip':
            if os.path.islink(path):
                # This is a symlink, we virtually create one in memory
                # because zipfile does not seem to support them natively
                vfile = zipfile.ZipInfo()
                vfile.filename = arcname  # That's the name of the actual file
                vfile.external_attr |= 0o120000 << long(16)  # symlink file type
                vfile.compress_type = zipfile.ZIP_STORED
                # os.readlink gives us the target of the symlink
                self.arch.writestr(vfile, os.readlink(path))
            else:
                self.arch.write(path, arcname)
        elif self.archive in ['tar.gz', 'tar.bz2']:
            self.arch.add(path, arcname=arcname, recursive=False)


def sanitize_string(string, strict=True):
    """Return a 'safe' version of the string (ie. remove malicious chars like
    '\n')

    :param string: String to escape
    :type string: str
    """
    if strict:
        return string.encode('unicode_escape').decode('utf-8')
    else:
        import re
        ret = repr(string).replace('\\\\', '\\')
        ret = re.sub(r"^u?'(.*)'$", r"\1", ret)
        return ret.decode('utf-8')


def lookup_file(name=None, guess=True, directory=False, check=True):
    if name and isinstance(name, basestring):
        if os.path.isfile(name) or name == '/dev/null':
            return name
        elif directory and os.path.isdir(name):
            return name
        elif not guess:
            if check:
                raise IOError('File not found: \'{}\''.format(name))
            return name
    if name and isinstance(name, basestring):
        names = [name]
    elif name:
        names = name
    else:
        names = ['burpui.cfg', 'burpui.sample.cfg']
    roots = [
        '',
        'share/burpui',
        '/etc/burp',
        os.path.join(
            sys.prefix,
            'share',
            'burpui',
        ),
        os.path.join(
            sys.prefix,
            'local',
            'share',
            'burpui',
        ),
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            '..',
            '..',
            '..',
            '..',
            'share',
            'burpui',
        ),
    ]
    prefixes = ['', 'etc']
    for filename in names:
        for root in roots:
            for prefix in prefixes:
                tmp = os.path.join(root, prefix, filename)
                if directory and os.path.isdir(tmp):
                    return tmp
                elif os.path.isfile(tmp):
                    return tmp


def utc_to_local(timestamp):
    try:
        import arrow
        from tzlocal import get_localzone
        utc = arrow.get(datetime.datetime.fromtimestamp(timestamp))
        local = utc.to(str(get_localzone()))
        return local.timestamp
    except (TypeError, arrow.parser.ParserError, ImportError):
        return timestamp


def implement(func):
    """A decorator indicating the method is implemented.

    For the agent and the 'multi' backend, we inherit the backend interface but
    we don't really implement it because we just act as a proxy.
    But maintaining the exhaustive list of methods in several places to always
    implement the same "proxy" thing was painful so I ended up cheating to
    dynamically implement those methods thanks to the __getattribute__ magic
    function.

    But sometimes we want to implement specific things, hence this decorator
    to indicate we don't want the default "magic" implementation and use the
    custom implementation instead.
    """
    func.__ismethodimplemented__ = True
    return func


def basic_login_from_request(request, app):
    """Check 'Authorization' headers and log the user in if possible.

    :param request: The input request
    :type request: :class:`flask.Request`

    :param app: The application context
    :type app: :class:`burpui.server.BUIServer`
    """
    if app.auth != 'none':
        if request.headers.get('X-From-UI', False):
            return None
        auth = request.authorization
        if auth:
            from flask import session, g
            app.logger.debug('Found Basic user: {}'.format(auth.username))
            refresh = False
            if 'login' in session and session['login'] != auth.username:
                refresh = True
                session.clear()
                session['login'] = auth.username
            user = app.uhandler.user(auth.username, refresh)
            if user.active and user.login(auth.password):
                from flask_login import login_user
                from .sessions import session_manager
                login_user(user)
                if request.headers.get('X-Reuse-Session', False):
                    session_manager.store_session(
                        auth.username,
                        request.remote_addr,
                        request.headers.get('User-Agent'),
                        remember=False,
                        api=True
                    )
                else:
                    g.basic_session = True
                app.logger.debug('Successfully logged in')
                return user
            app.logger.warning('Failed to log-in')
    return None


class ReverseProxied(object):
    '''Wrap the application in this middleware and configure the
    front-end server to add these headers, to let you quietly bind
    this to a URL other than / and to an HTTP scheme that is
    different than what is used locally.

    In nginx:

    ::

        location /myprefix {
            proxy_pass http://192.168.0.1:5001;
            proxy_set_header Host $host;
            proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
            proxy_set_header X-Scheme $scheme;
            proxy_set_header X-Script-Name /myprefix;
        }


    In Apache:

    ::

        <Location /myprefix>
            ProxyPass http://192.168.0.1:5001
            ProxyPassReverse http://192.168.0.1:5001
            RequestHeader set X-Script-Name /myprefix
        </Location>


    :param wsgi_app: the WSGI application

    Inspired by: http://flask.pocoo.org/snippets/35/
    '''
    def __init__(self, wsgi_app, app):
        self.wsgi_app = wsgi_app
        self.app = app

    def __call__(self, environ, start_response):
        script_name = environ.get('HTTP_X_SCRIPT_NAME', self.app.prefix)
        if script_name:
            if script_name.startswith('/'):
                environ['SCRIPT_NAME'] = script_name
                path_info = environ['PATH_INFO']
                if path_info.startswith(script_name):
                    environ['PATH_INFO'] = path_info[len(script_name):]
            else:
                self.app.warning("'prefix' must start with a '/'!")

        return self.wsgi_app(environ, start_response)



# -*- coding: utf8 -*-
"""
.. module:: burpui.ext.i18n
    :platform: Unix
    :synopsis: Burp-UI external Internationalization module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from flask import request
from flask_babel import Babel
from flask_login import current_user
from ..config import config

babel = Babel()

LANGUAGES = {
    'en': 'English',
    'fr': 'Français',
}
config['LANGUAGES'] = LANGUAGES


@babel.localeselector
def get_locale():
    locale = None
    if current_user and not current_user.is_anonymous:
        locale = getattr(current_user, 'language', None)
    return locale or request.accept_languages.best_match(config['LANGUAGES'].keys())



# -*- coding: utf8 -*-
"""
.. module:: burpui.exceptions
    :platform: Unix
    :synopsis: Burp-UI exceptions module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
# Agent do not need "real" HTTP errors
try:
    from werkzeug.exceptions import HTTPException
except ImportError:
    HTTPException = object


class BUIserverException(HTTPException):
    """Raised in case of internal error."""
    code = 500

    def __init__(self, message):
        self.description = message

    def __str__(self):
        return self.description



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.backup
    :platform: Unix
    :synopsis: Burp-UI backup api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from . import api
from ..server import BUIServer  # noqa
from .custom import Resource
from ..exceptions import BUIserverException

from flask import current_app

bui = current_app  # type: BUIServer
ns = api.namespace('backup', 'Backup methods')


@ns.route('/server-backup/<name>',
          '/<server>/server-backup/<name>',
          methods=['GET', 'DELETE'],
          endpoint='is_server_backup')
@ns.route('/do-server-backup/<name>',
          '/<server>/do-server-backup/<name>',
          methods=['PUT'],
          endpoint='server_backup')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'name': 'Client name',
    },
)
class ServerBackup(Resource):
    """The :class:`burpui.api.backup.ServerBackup` resource allows you to
    prepare a server-initiated backup.

    This resource is part of the :mod:`burpui.api.backup` module.
    """
    @ns.doc(
        responses={
            200: 'Success',
            400: 'Missing parameter',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    def get(self, server=None, name=None):
        """Tells if a 'backup' file is present

        **GET** method provided by the webservice.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :returns: True if the file is found
        """
        if not name:
            self.abort(400, 'Missing options')
        # Manage ACL
        if (bui.acl and
                (not bui.acl.is_client_allowed(self.username,
                                               name,
                                               server) and not
                 self.is_admin)):
            self.abort(403, 'You are not allowed to access this client')
        try:
            return {'is_server_backup': bui.client.is_server_backup(name, server)}
        except BUIserverException as e:
            self.abort(500, str(e))

    @ns.doc(
        responses={
            200: 'Success',
            400: 'Missing parameter',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    def delete(self, server=None, name=None):
        """Remove the 'backup' file if present

        **DELETE** method provided by the webservice.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :returns: Status message (success or failure)
        """
        if not name:
            self.abort(400, 'Missing options')
        # Manage ACL
        if (bui.acl and
                (not bui.acl.is_client_allowed(self.username,
                                               name,
                                               server) and not
                 self.is_admin)):
            self.abort(403, 'You are not allowed to cancel a backup for this client')
        try:
            return bui.client.cancel_server_backup(name, server)
        except BUIserverException as e:
            self.abort(500, str(e))

    @ns.doc(
        responses={
            201: 'Success',
            400: 'Missing parameter',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    def put(self, server=None, name=None):
        """Schedule a server-initiated backup

        **PUT** method provided by the webservice.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :returns: Status message (success or failure)
        """
        json = []
        # Check params
        if not name:
            self.abort(400, 'Missing options')
        # Manage ACL
        if (bui.acl and
                (not bui.acl.is_client_allowed(self.username,
                                               name,
                                               server) and not
                 self.is_admin)):
            self.abort(
                403,
                'You are not allowed to schedule a backup for this client'
            )
        try:
            json = bui.client.server_backup(name, server)
            return json, 201
        except BUIserverException as e:
            self.abort(500, str(e))



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.settings
    :platform: Unix
    :synopsis: Burp-UI settings api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from . import api
from ..server import BUIServer  # noqa
from .custom import Resource
from .custom.inputs import boolean
from .._compat import unquote
from ..utils import NOTIF_INFO

from six import iteritems
from flask_babel import gettext as _
from flask import jsonify, request, url_for, current_app
from werkzeug.datastructures import ImmutableMultiDict

bui = current_app  # type: BUIServer
ns = api.namespace('settings', 'Settings methods')


@ns.route('/server-config',
          '/<server>/server-config',
          '/server-config/<path:conf>',
          '/<server>/server-config/<path:conf>',
          endpoint='server_settings')
@ns.doc(
    params={
        'conf': 'Path of the configuration file',
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class ServerSettings(Resource):
    """The :class:`burpui.api.settings.ServerSettings` resource allows you to
    read and write the server's configuration.

    This resource is part of the :mod:`burpui.api.settings` module.
    """

    @api.disabled_on_demo()
    @api.acl_admin_required(message='Sorry, you don\'t have rights to access the setting panel')
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        }
    )
    def post(self, conf=None, server=None):
        """Saves the server configuration"""
        noti = bui.client.store_conf_srv(request.form, conf, server)
        return {'notif': noti}, 200

    @api.acl_admin_required(message='Sorry, you don\'t have rights to access the setting panel')
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        }
    )
    def get(self, conf=None, server=None):
        """Reads the server configuration

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "boolean": [
                "daemon",
                "fork",
                "..."
              ],
              "defaults": {
                "address": "",
                "autoupgrade_dir": "",
                "ca_burp_ca": "",
                "ca_conf": "",
                "ca_name": "",
                "ca_server_name": "",
                "client_can_delete": true,
                "...": "..."
              },
              "integer": [
                "port",
                "status_port",
                "..."
              ],
              "multi": [
                "keep",
                "restore_client",
                "..."
              ],
              "placeholders": {
                "autoupgrade_dir": "path",
                "ca_burp_ca": "path",
                "ca_conf": "path",
                "ca_name": "name",
                "ca_server_name": "name",
                "client_can_delete": "0|1",
                "...": "..."
              },
              "results": {
                "boolean": [
                  {
                    "name": "hardlinked_archive",
                    "value": false
                  },
                  {
                    "name": "syslog",
                    "value": true
                  },
                  { "...": "..." }
                ],
                "clients": [
                  {
                    "name": "testclient",
                    "value": "/etc/burp/clientconfdir/testclient"
                  }
                ],
                "common": [
                  {
                    "name": "mode",
                    "value": "server"
                  },
                  {
                    "name": "directory",
                    "value": "/var/spool/burp"
                  },
                  { "...": "..." }
                ],
                "includes": [],
                "includes_ext": [],
                "integer": [
                  {
                    "name": "port",
                    "value": 4971
                  },
                  {
                    "name": "status_port",
                    "value": 4972
                  },
                  { "...": "..." }
                ],
                "multi": [
                  {
                    "name": "keep",
                    "value": [
                      "7",
                      "4"
                    ]
                  },
                  { "...": "..." }
                ]
              },
              "server_doc": {
                "address": "Defines the main TCP address that the server listens on. The default is either '::' or '0.0.0.0', dependent upon compile time options.",
                "...": "..."
              },
              "string": [
                "mode",
                "address",
                "..."
              ],
              "suggest": {
                "compression": [
                  "gzip1",
                  "gzip2",
                  "gzip3",
                  "gzip4",
                  "gzip5",
                  "gzip6",
                  "gzip7",
                  "gzip8",
                  "gzip9"
                ],
                "mode": [
                  "client",
                  "server"
                ],
                "...": []
              }
            }


        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above.
        """
        try:
            conf = unquote(conf)
        except:
            pass
        r = bui.client.read_conf_srv(conf, server)
        # Translate the doc and placeholder API side
        doc = bui.client.get_parser_attr('doc', server)
        for key, val in iteritems(doc):
            doc[key] = _(val)
        placeholders = bui.client.get_parser_attr('placeholders', server)
        for key, val in iteritems(placeholders):
            placeholders[key] = _(val)
        return jsonify(results=r,
                       boolean=bui.client.get_parser_attr('boolean_srv', server),
                       string=bui.client.get_parser_attr('string_srv', server),
                       integer=bui.client.get_parser_attr('integer_srv', server),
                       multi=bui.client.get_parser_attr('multi_srv', server),
                       server_doc=doc,
                       suggest=bui.client.get_parser_attr('values', server),
                       placeholders=placeholders,
                       defaults=bui.client.get_parser_attr('defaults', server))


@ns.route('/clients',
          '/<server>/clients',
          endpoint='clients_list')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class ClientsList(Resource):

    @api.acl_admin_required(message='Sorry, you don\'t have rights to access the setting panel')
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        }
    )
    def get(self, server=None):
        """Returns a list of clients"""
        res = bui.client.clients_list(server)
        return jsonify(result=res)


@ns.route('/config',
          '/<server>/config',
          endpoint='new_client',
          methods=['PUT'])
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class NewClientSettings(Resource):
    parser = ns.parser()
    parser.add_argument('newclient', required=True, help="No 'newclient' provided")

    @api.disabled_on_demo()
    @api.acl_admin_required(message='Sorry, you don\'t have rights to access the setting panel')
    @ns.expect(parser)
    @ns.doc(
        responses={
            200: 'Success',
            400: 'Missing parameter',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        }
    )
    def put(self, server=None):
        """Creates a new client"""
        newclient = self.parser.parse_args()['newclient']
        if not newclient:
            self.abort(400, 'No client name provided')
        clients = bui.client.clients_list(server)
        for cl in clients:
            if cl['name'] == newclient:
                self.abort(409, "Client '{}' already exists".format(newclient))
        # clientconfdir = bui.client.get_parser_attr('clientconfdir', server)
        # if not clientconfdir:
        #    flash('Could not proceed, no \'clientconfdir\' find', 'warning')
        #    return redirect(request.referrer)
        noti = bui.client.store_conf_cli(ImmutableMultiDict(), newclient, None, server)
        if server:
            noti.append([NOTIF_INFO, '<a href="{}">Click here</a> to edit \'{}\' configuration'.format(url_for('view.cli_settings', server=server, client=newclient), newclient)])
        else:
            noti.append([NOTIF_INFO, '<a href="{}">Click here</a> to edit \'{}\' configuration'.format(url_for('view.cli_settings', client=newclient), newclient)])
        # clear the cache when we add a new client
        api.cache.clear()
        if bui.config['WITH_CELERY']:
            from .async import force_scheduling_now
            force_scheduling_now()
        return {'notif': noti}, 201


@ns.route('/config/<client>',
          '/config/<client>/<path:conf>',
          '/<server>/config/<client>',
          '/<server>/config/<client>/<path:conf>',
          endpoint='client_settings',
          methods=['GET', 'POST', 'DELETE'])
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'client': 'Client name',
        'conf': 'Path of the configuration file',
    },
)
class ClientSettings(Resource):
    parser_delete = ns.parser()
    parser_delete.add_argument('revoke', type=boolean, help='Whether to revoke the certificate or not', default=False, nullable=True)
    parser_delete.add_argument('delcert', type=boolean, help='Whether to delete the certificate or not', default=False, nullable=True)
    parser_delete.add_argument('keepconf', type=boolean, help='Whether to keep the conf or not', default=False, nullable=True)

    @api.disabled_on_demo()
    @api.acl_admin_required(message='Sorry, you don\'t have rights to access the setting panel')
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        }
    )
    def post(self, server=None, client=None, conf=None):
        """Saves a given client configuration"""
        noti = bui.client.store_conf_cli(request.form, client, conf, server)
        return {'notif': noti}

    @api.acl_admin_required(message='Sorry, you don\'t have rights to access the setting panel')
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        }
    )
    def get(self, server=None, client=None, conf=None):
        """Reads a given client configuration"""
        try:
            conf = unquote(conf)
        except:
            pass
        r = bui.client.read_conf_cli(client, conf, server)
        return jsonify(
            results=r,
            boolean=bui.client.get_parser_attr('boolean_cli', server),
            string=bui.client.get_parser_attr('string_cli', server),
            integer=bui.client.get_parser_attr('integer_cli', server),
            multi=bui.client.get_parser_attr('multi_cli', server),
            server_doc=bui.client.get_parser_attr('doc', server),
            suggest=bui.client.get_parser_attr('values', server),
            placeholders=bui.client.get_parser_attr('placeholders', server),
            defaults=bui.client.get_parser_attr('defaults', server)
        )

    @api.disabled_on_demo()
    @api.acl_admin_required(message='Sorry, you don\'t have rights to access the setting panel')
    @ns.expect(parser_delete)
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        }
    )
    def delete(self, server=None, client=None, conf=None):
        """Deletes a given client"""
        args = self.parser_delete.parse_args()
        delcert = args.get('delcert', False)
        revoke = args.get('revoke', False)
        keepconf = args.get('keepconf', False)

        if not keepconf:
            # clear the cache when we remove a client
            api.cache.clear()
            if bui.config['WITH_CELERY']:
                from .async import force_scheduling_now
                force_scheduling_now()
        return bui.client.delete_client(client, keepconf=keepconf, delcert=delcert, revoke=revoke, agent=server), 200


@ns.route('/path-expander',
          '/<server>/path-expander',
          '/path-expander/<client>',
          '/<server>/path-expander/<client>',
          endpoint='path_expander')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'client': 'Client name',
    },
)
class PathExpander(Resource):

    parser = ns.parser()
    parser.add_argument('path', required=True, help="No 'path' provided")
    parser.add_argument('source', required=False, help="Which file is it included in")

    @api.acl_admin_required(message='Sorry, you don\'t have rights to access the setting panel')
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        }
    )
    def get(self, server=None, client=None):
        """Expends a given path

        For instance if it's given a glob expression it will returns a list of
        files matching the expression.
        """
        args = self.parser.parse_args()
        path = args['path']
        source = args['source']
        if path:
            path = unquote(path)
        if source:
            source = unquote(source)
        paths = bui.client.expand_path(path, source, client, server)
        if not paths:
            self.abort(403, 'Path not found')
        return {'result': paths}


@ns.route('/options',
          '/<server>/options',
          endpoint='setting_options')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class SettingOptions(Resource):

    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        }
    )
    def get(self, server=None):
        """Returns various setting options"""
        return {
            'is_revocation_enabled': bui.client.revocation_enabled(server),
            'server_can_restore': not bui.noserverrestore or bui.client.get_parser(agent=server).param('server_can_restore', 'client_conf'),
            'batch_list_supported': bui.client.get_attr('batch_list_supported', False, server),
        }



# -*- coding: utf8 -*-

# This is a submodule we can also use "from ..api import api"
from . import api, cache_key
from ..server import BUIServer  # noqa
from .custom import fields, Resource
from ..exceptions import BUIserverException

from flask import current_app

bui = current_app  # type: BUIServer
ns = api.namespace('servers', 'Servers methods')


@ns.route('/stats', endpoint='servers_stats')
class ServersStats(Resource):
    """The :class:`burpui.api.servers.ServersStats` resource allows you to
    retrieve statistics about servers/agents.

    This resource is part of the :mod:`burpui.api.servers` module.
    """
    servers_fields = ns.model('Servers', {
        'alive': fields.Boolean(required=True, description='Is the server reachable'),
        'clients': fields.Integer(required=True, description='Number of clients managed by this server'),
        'name': fields.String(required=True, description='Server name'),
    })

    @api.cache.cached(timeout=1800, key_prefix=cache_key)
    @ns.marshal_list_with(servers_fields, code=200, description='Success')
    @ns.doc(
        responses={
            500: 'Internal failure',
        },
    )
    def get(self):
        """Returns a list of servers (agents) with basic stats

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                'alive': true,
                'clients': 2,
                'name': 'burp1',
              },
              {
                'alive': false,
                'clients': 0,
                'name': 'burp2',
              },
            ]


        :returns: The *JSON* described above.
        """

        r = []
        restrict = []
        check = False

        if bui.standalone:
            return r

        if bui.acl and not self.is_admin:
            check = True
            restrict = bui.acl.servers(self.username)

        for serv in bui.client.servers:
            try:
                alive = bui.client.servers[serv].ping()
            except BUIserverException:
                alive = False

            if check and serv in restrict:
                r.append({
                    'name': serv,
                    'clients': len(bui.acl.clients(self.username, serv)),
                    'alive': alive
                })
            elif not check:
                try:
                    clients = bui.client.servers[serv].get_all_clients(serv)
                except BUIserverException:
                    clients = []

                r.append({
                    'name': serv,
                    'clients': len(clients),
                    'alive': alive
                })

        return r


@ns.route('/report', endpoint='servers_report')
class ServersReport(Resource):
    """The :class:`burpui.api.servers.ServersReport` resource allows you to
    retrieve a report about servers/agents.

    This resource is part of the :mod:`burpui.api.servers` module.
    """
    stats_fields = ns.model('ServersStats', {
        'total': fields.Integer(required=True, description='Number of files', default=0),
        'totsize': fields.Integer(required=True, description='Total size occupied by all the backups of this server', default=0),
        'linux': fields.Integer(required=True, description='Total number of Linux/Unix clients on this server', default=0),
        'windows': fields.Integer(required=True, description='Total number of Windows clients on this server', default=0),
        'unknown': fields.Integer(required=True, description='Total number of Unknown clients on this server', default=0),
    })
    server_fields = ns.model('ServersReport', {
        'name': fields.String(required=True, description='Server name'),
        'stats': fields.Nested(stats_fields, required=True),
    })
    backup_fields = ns.model('ServersBackup', {
        'name': fields.String(required=True, description='Server name'),
        'number': fields.Integer(required=True, description='Number of backups on this server', default=0),
    })
    report_fields = ns.model('ServersReportFull', {
        'backups': fields.Nested(backup_fields, as_list=True, required=True),
        'servers': fields.Nested(server_fields, as_list=True, required=True),
    })

    @api.cache.cached(timeout=1800, key_prefix=cache_key)
    @ns.marshal_with(report_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    def get(self):
        """Returns a global report about all the servers managed by Burp-UI

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "backups": [
                {
                  "name": "AGENT1",
                  "number": 49
                }
              ],
              "servers": [
                {
                  "name": "AGENT1",
                  "stats": {
                    "linux": 4,
                    "total": 349705,
                    "totsize": 119400711726,
                    "unknown": 0,
                    "windows": 1
                  }
                }
              ]
            }

        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients/servers you are authorized to.

        :returns: The *JSON* described above.
        """
        r = {}
        restrict = []
        check = False
        if bui.acl and not self.is_admin:
            check = True
            restrict = bui.acl.servers(self.username)

        backups = []
        servers = []
        try:
            for serv in bui.client.servers:
                out = {
                    'name': serv,
                    'stats': {
                        'total': 0,
                        'totsize': 0,
                        'linux': 0,
                        'windows': 0,
                        'unknown': 0,
                    },
                    'number': 0
                }
                if check and serv not in restrict:
                    continue
                clients = []
                if bui.acl and not self.is_admin:
                    clients = [{'name': x} for x in bui.acl.clients(self.username, serv)]
                else:
                    clients = bui.client.get_all_clients(agent=serv)

                j = bui.client.get_clients_report(clients, serv)
                if 'clients' not in j or 'backups' not in j:
                    continue
                for stats in j['clients']:
                    for key in ['total', 'totsize']:
                        out['stats'][key] += stats['stats'][key]
                    if stats['stats']['windows'] == 'true':
                        out['stats']['windows'] += 1
                    elif stats['stats']['windows'] == 'false':
                        out['stats']['linux'] += 1
                    else:
                        out['stats']['unknown'] += 1
                for bkp in j['backups']:
                    out['number'] += bkp['number']
                backups.append({'name': serv, 'number': out['number']})
                servers.append({'name': serv, 'number': out['stats']})

        except BUIserverException as e:
            self.abort(500, str(e))

        r['backups'] = backups
        r['servers'] = servers

        return r



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.parser.doc
    :platform: Unix
    :synopsis: Burp-UI parser documentation for Burp1.
.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
from .interface import BUIparser


def __(string):
    """dummy function to fake the translation"""
    return string


class Doc(BUIparser):
    """:class:`burpui.misc.parser.doc.Doc` provides a consistent interface
    to parse burp configuration files.

    It implements :class:`burpui.misc.parser.interface.BUIparser`.
    """
    defaults = {
        u'address': u'',  # IP
        u'atime': False,  # bool
        u'autoupgrade_dir': u'',  # dir
        u'ca_burp_ca': u'',  # file
        u'ca_conf': u'',  # file
        u'ca_name': u'',
        u'ca_server_name': u'',
        u'client_can_delete': True,  # bool
        u'client_can_diff': True,  # bool
        u'client_can_force_backup': True,  # bool
        u'client_can_list': True,  # bool
        u'client_can_restore': True,  # bool
        u'client_can_verify': True,  # bool
        u'client_lockdir': u'',
        u'clientconfdir': u'',  # dir
        u'compression': u'gzip9',
        u'cross_all_filesystems': False,  # bool
        u'cross_filesystem': u'',
        u'daemon': True,  # bool
        u'dedup_group': u'',
        u'directory_tree': True,  # bool
        u'directory': u'',
        u'exclude_comp': u'',  # multi
        u'exclude_ext': u'',  # multi
        u'exclude_fs': u'',  # multi
        u'exclude_regex': u'',  # multi
        u'exclude': u'',  # multi
        u'fork': True,  # bool
        u'group': u'',
        u'hard_quota': u'',
        u'hardlinked_archive': False,  # bool
        u'include_ext': u'',  # multi
        u'include_glob': u'',  # multi
        u'include_regex': u'',  # multi
        u'include': u'',  # multi
        u'keep': [7, 6, 4],  # multi  # int
        u'librsync': True,  # bool
        u'max_children': 5,  # int
        u'max_file_size': u'',
        u'max_hardlinks': 10000,  # int
        u'max_status_children': 5,  # int
        u'max_storage_subdirs': 30000,  # int
        u'min_file_size': u'',
        u'mode': u'',
        u'monitor_browse_cache': False,  # bool
        u'network_timeout': 7200,  # int
        u'nobackup': u'',
        u'notify_failure_arg': u'',  # multi
        u'notify_failure_script': u'',  # file
        u'notify_success_arg': u'',  # multi
        u'notify_success_changes_only': False,  # bool
        u'notify_success_script': u'',  # file
        u'notify_success_warnings_only': False,  # bool
        u'password': u'',
        u'password_check': True,  # bool
        u'path_length_warn': True,  # bool
        u'pidfile': u'',
        u'port': 4971,  # int
        u'protocol': False,  # int
        u'ratelimit': False,  # int
        u'read_all_blockdevs': False,  # bool
        u'read_all_fifos': False,  # bool
        u'read_blockdev': u'',
        u'read_fifo': u'',
        u'restore_client': u'',  # multi
        u'scan_problem_raises_error': False,  # bool
        u'server_script_arg': u'',  # multi
        u'server_script_notify': False,  # bool
        u'server_script_post_arg': u'',  # multi
        u'server_script_post_notify': False,  # bool
        u'server_script_post_run_on_fail': False,  # bool
        u'server_script_post': u'',  # file
        u'server_script_pre_arg': u'',  # multi
        u'server_script_pre_notify': False,  # bool
        u'server_script_pre': u'',  # file
        u'server_script': u'',  # file
        u'soft_quota': u'',
        u'split_vss': False,  # bool
        u'ssl_cert_ca': u'',  # file
        u'ssl_cert': u'',  # file
        u'ssl_ciphers': u'',
        u'ssl_compression': u'zlib5',
        u'ssl_dhfile': u'',  # file
        u'ssl_key_password': u'',
        u'ssl_key': u'',  # file
        u'ssl_peer_cn': u'',
        u'status_address': u'',  # 127.0.0.1 / ::1
        u'status_port': 4972,  # int
        u'stdout': True,  # bool
        u'strip_vss': False,  # bool
        u'syslog': False,  # bool
        u'timer_arg': u'',  # multi
        u'timer_script': u'',  # file
        u'timestamp_format': u'',
        u'umask': u'0022',  # mode
        u'user': u'',
        u'version_warn': True,  # bool
        u'vss_drives': u'',
        u'working_dir_recovery_method': u'',
        u'server_can_restore': False,
    }
    placeholders = {
        u'.': __(u"path or glob"),
        u'atime': u"0|1",
        u'autoupgrade_dir': __(u"path"),
        u'ca_burp_ca': __(u"path"),
        u'ca_conf': __(u"path"),
        u'ca_name': __(u"name"),
        u'ca_server_name': __(u"name"),
        u'client_can_delete': u"0|1",
        u'client_can_force_backup': u"0|1",
        u'client_can_list': u"0|1",
        u'client_can_restore': u"0|1",
        u'client_can_verify': u"0|1",
        u'client_lockdir': __(u"path"),
        u'clientconfdir': __(u"path"),
        u'compression': u"gzip[0-9]",
        u'cross_all_filesystems': u"0|1",
        u'cross_filesystem': __(u"path"),
        u'daemon': u"0|1",
        u'dedup_group': __(u"string"),
        u'directory_tree': u"0|1",
        u'directory': __(u"path"),
        u'exclude_comp': __(u"extension"),
        u'exclude_ext': __(u"extension"),
        u'exclude_fs': __(u"fstype"),
        u'exclude_regex': __(u"regular expression"),
        u'exclude': __(u"path"),
        u'fork': u"0|1",
        u'group': __(u"groupname"),
        u'hard_quota': u"b/Kb/Mb/Gb",
        u'hardlinked_archive': u"0|1",
        u'include_ext': __(u"extension"),
        u'include_regex': __(u"regular expression"),
        u'include': __(u"path"),
        u'include_glob': __(u"glob"),
        u'keep': __(u"number"),
        u'librsync': u"0|1",
        u'lockfile': __(u"path"),
        u'manual_delete': __(u"path"),
        u'max_children': __(u"number"),
        u'max_file_size': u"b/Kb/Mb/Gb",
        u'max_hardlinks': __(u"number"),
        u'max_status_children': __(u"number"),
        u'max_storage_subdirs': __(u"number"),
        u'min_file_size': u"b/Kb/Mb/Gb",
        u'network_timeout': u"s",
        u'nobackup': __(u"file name"),
        u'notify_failure_arg': __(u"string"),
        u'notify_failure_script': __(u"path"),
        u'notify_success_arg': __(u"string"),
        u'notify_success_changes_only': u"0|1",
        u'notify_success_script': __(u"path"),
        u'notify_success_warnings_only': u"0|1",
        u'password': __(u"password"),
        u'password_check': u"0|1",
        u'pidfile': __(u"path"),
        u'port': __(u"port number"),
        u'ratelimit': u"Mb/s",
        u'read_all_blockdevs': u"0|1",
        u'read_all_fifos': u"0|1",
        u'read_blockdev': __(u"path"),
        u'read_fifo': __(u"path"),
        u'restore_client': __(u"client"),
        u'resume_partial': u"0|1",
        u'scan_problem_raises_error': u"0|1",
        u'server_script_arg': __(u"path"),
        u'server_script_notify': u"0|1",
        u'server_script_post_arg': __(u"string"),
        u'server_script_post_notify': u"0|1",
        u'server_script_post_run_on_fail': u"0|1",
        u'server_script_post': __(u"path"),
        u'server_script_pre_arg': __(u"string"),
        u'server_script_pre_notify': u"0|1",
        u'server_script_pre': __(u"path"),
        u'server_script': __(u"path"),
        u'soft_quota': u"b/Kb/Mb/Gb",
        u'ssl_cert_ca': __(u"path"),
        u'ssl_cert_password': __(u"password"),
        u'ssl_cert': __(u"path"),
        u'ssl_ciphers': __(u"cipher list"),
        u'ssl_dhfile': __(u"path"),
        u'ssl_key_password': __(u"password"),
        u'ssl_key': __(u"path"),
        u'ssl_peer_cn': __(u"string"),
        u'status_port': __(u"port number"),
        u'stdout': u"0|1",
        u'strip_vss': u"0|1",
        u'suplit_vss': u"0|1",
        u'syslog': u"0|1",
        u'timer_arg': __(u"string"),
        u'timer_script': __(u"path"),
        u'timestamp_format': __(u"strftime format"),
        u'umask': __(u"umask"),
        u'user': __(u"username"),
        u'version_warn': u"0|1",
        u'vss_drives': __(u"list of drive letters"),
        u'working_dir_recovery_method': u"resume|use|delete",
    }
    values = {
        u'compression': [u'gzip{0}'.format(x) for x in range(0, 10)],
        u'mode': [u'client', u'server'],
        u'ssl_compression': [u'zlib{0}'.format(x) for x in range(0, 10)],
        u'status_address': [u'127.0.0.1', u'::1'],  # 127.0.0.1 / ::1
        u'working_dir_recovery_method': [u'use', u'delete', u'resume'],
    }
    files = [
        u'ca_burp_ca',
        u'ca_conf',
        u'notify_failure_script',
        u'notify_success_script',
        u'server_script_post',
        u'server_script_pre',
        u'server_script',
        u'ssl_cert_ca',
        u'ssl_cert',
        u'ssl_dhfile',
        u'ssl_key',
        u'timer_script',
    ]
    multi_srv = [
        u'exclude_comp',
        u'exclude_ext',
        u'exclude_fs',
        u'exclude_regex',
        u'exclude',
        u'include_ext',
        u'include_glob',
        u'include_regex',
        u'include',
        u'keep',
        u'notify_failure_arg',
        u'notify_success_arg',
        u'restore_client',
        u'server_script_arg',
        u'server_script_post_arg',
        u'server_script_pre_arg',
        u'timer_arg',
    ]
    boolean_srv = [
        u'atime',
        u'client_can_delete',
        u'client_can_diff',
        u'client_can_force_backup',
        u'client_can_list',
        u'client_can_restore',
        u'client_can_verify',
        u'cross_all_filesystems',
        u'daemon',
        u'directory_tree',
        u'fork',
        u'hardlinked_archive',
        u'librsync',
        u'monitor_browse_cache',
        u'notify_success_changes_only',
        u'notify_success_warnings_only',
        u'password_check',
        u'path_length_warn',
        u'read_all_blockdevs',
        u'read_all_fifos',
        u'scan_problem_raises_error',
        u'server_script_notify',
        u'server_script_post_notify',
        u'server_script_post_run_on_fail',
        u'server_script_pre_notify',
        u'split_vss',
        u'stdout',
        u'strip_vss',
        u'syslog',
        u'version_warn',
        u'server_can_restore'
    ]
    integer_srv = [
        u'max_children',
        u'max_hardlinks',
        u'max_status_children',
        u'max_storage_subdirs',
        u'network_timeout',
        u'port',
        u'protocol',
        u'ratelimit',
        u'status_port',
    ]
    string_srv = [
        u'address',
        u'ca_burp_ca',
        u'ca_conf',
        u'ca_name',
        u'ca_server_name',
        u'client_lockdir',
        u'compression',
        u'dedup_group',
        u'directory',
        u'group',
        u'hard_quota',
        u'mode',
        u'notify_failure_script',
        u'notify_success_script',
        u'pidfile',
        u'server_script_post',
        u'server_script_pre',
        u'server_script',
        u'soft_quota',
        u'ssl_cert_ca',
        u'ssl_cert',
        u'ssl_ciphers',
        u'ssl_compression',
        u'ssl_dhfile',
        u'ssl_key_password',
        u'ssl_key',
        u'status_address',
        u'timestamp_format',
        u'umask',
        u'user',
        u'working_dir_recovery_method',
        u'min_file_size',
        u'max_file_size',
        u'cross_filesystem',
        u'nobackup',
        u'read_fifo',
        u'read_blockdev',
        u'vss_drives',
    ]
    fields_cli = [
        u'atime',
        u'client_can_delete',
        u'client_can_force_backup',
        u'client_can_list',
        u'client_can_restore',
        u'client_can_verify',
        u'compression',
        u'cross_all_filesystems',
        u'cross_filesystem',
        u'dedup_group',
        u'directory_tree',
        u'directory',
        u'exclude_comp',
        u'exclude_ext',
        u'exclude_fs',
        u'exclude_regex',
        u'exclude',
        u'hard_quota',
        u'include_ext',
        u'include_regex',
        u'include_glob',
        u'include',
        u'keep',
        u'librsync',
        u'max_file_size',
        u'min_file_size',
        u'nobackup',
        u'notify_failure_arg',
        u'notify_failure_script'
        u'notify_success_arg',
        u'notify_success_script',
        u'notify_success_warnings_only',
        u'password_check',
        u'password',
        u'path_length_warn',
        u'protocol',
        u'read_all_blockdevs',
        u'read_all_fifos',
        u'read_blockdev',
        u'read_fifo',
        u'restore_client',
        u'scan_problem_raises_error',
        u'server_script_arg',
        u'server_script_notify',
        u'server_script_post_arg',
        u'server_script_post_notify',
        u'server_script_post_run_on_fail',
        u'server_script_post',
        u'server_script_pre_arg',
        u'server_script_pre_notify',
        u'server_script_pre',
        u'server_script',
        u'soft_quota',
        u'split_vss',
        u'ssl_peer_cn',
        u'strip_vss',
        u'syslog',
        u'timer_arg',
        u'timer_script',
        u'timestamp_format',
        u'version_warn',
        u'vss_drives',
        u'working_dir_recovery_method',
        u'server_can_restore',
    ]
    string_cli = list(set(string_srv) & set(fields_cli))
    string_cli += [u'ssl_peer_cn', u'password']
    boolean_cli = list(set(boolean_srv) & set(fields_cli))
    integer_cli = list(set(integer_srv) & set(fields_cli))
    multi_cli = list(set(multi_srv) & set(fields_cli))
    doc = {
        u'.': __(u"Read additional configuration files. On Windows, the glob"
                 " is unimplemented - you will need to specify an actual"
                 " file."),
        u'address': __(u"Defines the main TCP address that the server listens"
                       " on. The default is either '::' or '0.0.0.0',"
                       " dependent upon compile time options."),
        u'atime': __(u"This allows you to control whether the client uses"
                     " O_NOATIME when opening files and directories. The"
                     " default is 0, which enables O_NOATIME. This means that"
                     " the client can read files and directories without"
                     " updating the access times. However, this is only"
                     " possible if you are running as root, or are the owner"
                     " of the file or directory. If this is not the case"
                     " (perhaps you only have group or world access to the"
                     " files), you will get errors until you set atime=1."
                     " With atime=1, the access times will be updated on the"
                     " files and directories that get backed up."),
        u'autoupgrade_dir': __(u"Path to autoupgrade directory from which"
                               " upgrades are downloaded. The option can be"
                               " left unset in order not to autoupgrade"
                               " clients. Please see docs/autoupgrade.txt in"
                               " the source package for more help with this"
                               " option."),
        u'ca_burp_ca': __(u"Path to the burp_ca script when using the ca_conf"
                          " option."),
        u'ca_conf': __(u"Path to certificate authority configuration file. The"
                       " CA configuration file will usually be"
                       " /etc/burp/CA.cnf. The CA directory indicated by"
                       " CA.cnf will usually be /etc/burp/CA. If ca_conf is"
                       " set and the CA directory does not exist, the server"
                       " will create, populate it, and the paths indicated by"
                       " ssl_cert_ca, ssl_cert, ssl_key and ssl_dhfile will be"
                       " overwritten. For more detailed information on this"
                       " and the other ca_* options, please see"
                       " docs/burp_ca.txt."),
        u'ca_name': __(u"Name of the CA that the server will generate when"
                       " using the ca_conf option."),
        u'ca_server_name': __(u"The name that the server will put into its own"
                              " SSL certficates when using the ca_conf"
                              " option."),
        u'client_can_delete': __(u"Turn this off to prevent clients from"
                                 " deleting backups with the '-a D' option."
                                 " The default is that clients can delete"
                                 " backups. Restore clients can override this"
                                 " setting."),
        u'client_can_force_backup': __(u"Turn this off to prevent clients from"
                                       " forcing backups with the '-a b'"
                                       " option. Timed backups will still"
                                       " work. The default is that clients can"
                                       " force backups."),
        u'client_can_list': __(u"Turn this off to prevent clients from listing"
                               " backups with the '-a l' option. The default"
                               " is that clients can list backups. Restore"
                               " clients can override this setting."),
        u'client_can_restore': __(u"Turn this off to prevent clients from"
                                  " initiating restores with the '-a r'"
                                  " option. The default is that clients can"
                                  " initiate restores. Restore clients can"
                                  " override this setting."),
        u'client_can_verify': __(u"Turn this off to prevent clients from"
                                 " initiating a verify job with the '-a v'"
                                 " option. The default is that clients can"
                                 " initiate a verify job. Restore clients can"
                                 " override this setting."),
        u'client_lockdir': __(u"Path to the directory in which to keep"
                              " per-client lock files. By default, this is set"
                              " to the path given by the 'directory' option."),
        u'clientconfdir': __(u"Path to the directory that contains client"
                             " configuration files."),
        u'compression': __(u"Choose the level of gzip compression for files"
                           " stored in backups. Setting 0 or gzip0 turns"
                           " compression off. The default is gzip9. This"
                           " option can be overridden by the client"
                           " configuration files in clientconfdir on the"
                           " server."),
        u'cross_all_filesystems': __(u"Allow backups to cross all filesystem"
                                     " mountpoints."),
        u'cross_filesystem': __(u"Allow backups to cross a particular"
                                " filesystem mountpoint."),
        u'daemon': __(u"Whether to daemonise. The default is 1."),
        u'dedup_group': __(u"Enables you to group clients together for file"
                           " deduplication purposes. For example, you might"
                           " want to set 'dedup_group=xp' for each Windows XP"
                           " client, and then run the bedup program on a cron"
                           " job every other day with the option '-g xp'."),
        u'directory_tree': __(u"When turned on (which is the default) and the"
                              " client is on version 1.3.6 or greater, the"
                              " structure of the storage directory will mimic"
                              " that of the original filesystem on the"
                              " client."),
        u'directory': __(u"Path to the directory in which to store backups."),
        u'exclude_comp': __(u"Extensions to exclude from compression. Case"
                            " insensitive. You can have multiple exclude"
                            " compression lines. For example, set 'gz' to"
                            " exclude gzipped files from compression."),
        u'exclude_ext': __(u"Extensions to exclude from the backup. Case"
                           " insensitive. You can have multiple exclude"
                           " extension lines. For example, set 'vdi' to"
                           " exclude VirtualBox disk images."),
        u'exclude_fs': __(u"File systems to exclude from the backup. Case"
                          " insensitive. You can have multiple exclude file"
                          " system lines. For example, set 'tmpfs' to exclude"
                          " tmpfs. Burp has an internal mapping of file system"
                          " names to file system IDs. If you know the file"
                          " system ID, you can use that instead. For example,"
                          " 'exclude_fs = 0x01021994' will also"
                          " exclude tmpfs."),
        u'exclude_regex': __(u"Exclude paths that match the regular"
                             " expression."),
        u'exclude': __(u"Path to exclude from the backup. You can have"
                       " multiple exclude lines. Use forward slashes '/', not"
                       " backslashes '\\' as path delimiters."),
        u'fork': __(u"Whether to fork children. The default is 1."),
        u'group': __(u"Run as a particular group. This can be overridden by"
                     " the client configuration files in clientconfdir on the"
                     " server."),
        u'hard_quota': __(u"Do not back up the client if the estimated size of"
                          " all files is greater than the specified size."
                          " Example: 'hard_quota = 100Gb'. Set to 0 (the"
                          " default) to have no limit."),
        u'hardlinked_archive': __(u"On the server, defines whether to keep"
                                  " hardlinked files in the backups, or"
                                  " whether to generate reverse deltas and"
                                  " delete the original files. Can be set to"
                                  " either 0 (off) or 1 (on). Disadvantage:"
                                  " More disk space will be used Advantage:"
                                  " Restores will be faster, and since no"
                                  " reverse deltas need to be generated, the"
                                  " time and effort the server needs at the"
                                  " end of a backup is reduced."),
        u'include_ext': __(u"Extensions to include in the backup. Case"
                           " insensitive. Nothing else will be included in the"
                           " backup. You can have multiple include extension"
                           " lines. For example, set 'txt' to include files"
                           " that end in '.txt'. You need to specify an"
                           " 'include' line so that burp knows where to start"
                           " looking."),
        u'include_regex': __(u"Not implemented."),
        u'include': __(u"Path to include in the backup. You can have multiple"
                       " include lines. Use forward slashes '/', not"
                       " backslashes '\\' as path delimiters."),
        u'include_glob': __(u"Include paths that match the glob expression."
                            "For example, '/home/*/Documents' will include"
                            " '/home/user1/Documents' and"
                            " '/home/user2/Documents' if directories 'user1'"
                            " and 'user2' exist in '/home'. The Windows"
                            " implementation currently limit the expression to"
                            " contain only one '*'."),
        u'keep': __(u"Number of backups to keep. This can be overridden by the"
                    " client configuration files in clientconfdir on the"
                    " server. Specify multiple 'keep' entries on separate"
                    " lines in order to keep multiple periods of backups. For"
                    " example, assuming that you are doing a backup a day,"
                    " keep=7 keep=4 keep=6 (on separate lines) will keep 7"
                    " daily backups, 4 weekly backups (7x4=28), and 6"
                    " multiples of 4 weeks (7x4x6=168) - roughly 6 monthly"
                    " backups. Effectively, you will be guaranteed to be able"
                    " to restore up to 168 days ago, with the number of"
                    " available backups exponentially decreasing as you go"
                    " back in time to that point. In this example, every 7th"
                    " backup will be hardlinked to allow burp to safely delete"
                    " intermediate backups when necessary. You can have as"
                    " many 'keep' lines as you like, as long as they don't"
                    " exceed 52560000 when multiplied together. That is, a"
                    " backup every minute for 100 years."),
        u'librsync': __(u"When set to 0, delta differencing will not take"
                        " place. That is, when a file changes, the server will"
                        " request the whole new file. The default is 1. This"
                        " option can be overridden by the client configuration"
                        " files in clientconfdir on the server."),
        u'lockfile': __(u"Path to the lockfile that ensures that two server"
                        " processes cannot run simultaneously."),
        u'manual_delete': __(u"If a path is given, the server will move"
                             " directories to be deleted into the directory"
                             " specified by the path, but will not actually"
                             " delete them. The path must be on the same file"
                             " system as the backup storage. The idea is that"
                             " a busy server may be configured to run the"
                             " deletions outside of the backup timebands, when"
                             " the server is less busy, via a cron job. The"
                             " default is unset, which means that the server"
                             " will automatically delete the directories at"
                             " the end of a backup. This option can be"
                             " overridden by the client configuration files in"
                             " clientconfdir on the server."),
        u'max_children': __(u"Defines the number of child processes to fork"
                            " (the number of clients that can simultaneously"
                            " connect. The default is 5."),
        u'max_file_size': __(u"Do not back up files that are greater than the"
                             " specified size. Example: 'max_file_size ="
                             " 10Mb'. Set to 0 (the default) to have no"
                             " limit."),
        u'max_hardlinks': __("On the server, the number of times that a single"
                             " file can be hardlinked. The bedup program also"
                             " obeys this setting. The default is 10000."),
        u'max_status_children': __(u"Defines the number of status child"
                                   " processes to fork (the number of status"
                                   " clients that can simultaneously connect."
                                   " The default is 5."),
        u'max_storage_subdirs': __(u"Defines the number of subdirectories in"
                                   " the data storage areas. The maximum number"
                                   " of subdirectories that ext3 allows is"
                                   " 32000. If you do not set this option, it"
                                   " defaults to 30000."),
        u'min_file_size': __(u"Do not back up files that are less than the"
                             " specified size. Example: 'min_file_size ="
                             " 10Mb'. Set to 0 (the default) to have no"
                             " limit."),
        u'mode': __(u"Required to run in server mode."),
        u'monitor_browse_cache': __(u"Whether or not the server should cache"
                                    " the directory tree when a monitor client"
                                    " is browsing. <br/>Advantage: browsing is"
                                    " faster. </br>Disadvantage: more memory is"
                                    " used."),
        u'network_timeout': __(u"Set the network timeout in seconds. If no"
                               " data is sent or received over a period of"
                               " this length, burp will give up. The default"
                               " is 7200 seconds (2 hours)."),
        u'nobackup': __(u"If this file system entry exists, the directory"
                        " containing it will not be backed up."),
        u'notify_failure_arg': __(u"The same as notify_success_arg, but for"
                                  " backups that failed."),
        u'notify_failure_script': __(u"The same as notify_success_script, but"
                                     " for backups that failed."),
        u'notify_success_arg': __(u"A user-definable argument to the notify"
                                  " success script. You can have many of"
                                  " these. The notify_success_arg options can"
                                  " be overridden by the client configuration"
                                  " files in clientconfdir on the server."),
        u'notify_success_script': __(u"Path to the script to run when a backup"
                                     " succeeds. User arguments are appended"
                                     " after the first five reserved"
                                     " arguments. An example notify script is"
                                     " provided. The notify_success_script"
                                     " option can be overridden by the client"
                                     " configuration files in clientconfdir on"
                                     " the server."),
        u'notify_success_warnings_only': __(u"Set to 1 to send success"
                                            " notifications when there were"
                                            " warnings. If this and"
                                            " notify_success_changes_only are"
                                            " not turned on, success"
                                            " notifications are always sent."),
        u'password': __(u"Defines the password to send to the server."),
        u'password_check': __(u"Allows you to turn client password checking on"
                              " or off. The default is on. SSL certificates"
                              " will still be checked if you turn passwords"
                              " off. This option can be overridden by the"
                              " client configuration files in clientconfdir on"
                              " the server."),
        u'path_length_warn': __(u"When this is on, which is the default, a"
                                " warning will be issued when the client sends"
                                " a path that is too long to replicate in the"
                                " storage area tree structure. The file will"
                                " still be saved in a numbered file outside of"
                                " the tree structure, regardless of the"
                                " setting of this option. This option can be"
                                " overridden by the client configuration files"
                                " in clientconfdir on the server."),
        u'pidfile': __(u"Synonym for lockfile."),
        u'port': __(u"Defines the main TCP port that the server listens on."),
        u'protocol': __(u"Choose which style of backups and restores to use. 0"
                        " (the default) automatically decides based on the"
                        " server version and which protocol is set on the"
                        " server side. 1 forces protocol1 style (file level"
                        " granularity with a pseudo mirrored storage on the"
                        " server and optional rsync). 2 forces protocol2 style"
                        " (inline deduplication with variable length blocks)."
                        " If you choose a forced setting, it will be an error"
                        " if the server also chooses a forced setting."),
        u'ratelimit': __(u"Set the network send rate limit, in Mb/s. If this"
                         " option is not given, burp will send data as fast as"
                         " it can."),
        u'read_all_blockdevs': __(u"Open all block devices for reading and"
                                  " back up the contents as if they were"
                                  " regular files."),
        u'read_all_fifos': __(u"Open all fifos for reading and back up the"
                              " contents as if they were regular files."),
        u'read_blockdev': __(u"Do not back up the given block device itself,"
                             " but open it for reading and back up the"
                             " contents as if it were a regular file."),
        u'read_fifo': __(u"Do not back up the given fifo itself, but open it"
                         " for reading and back up the contents as if it were"
                         " a regular file."),
        u'restore_client': __(u"A client that is permitted to list, verify,"
                              " restore and delete files belonging to any"
                              " other client. You may specify multiple"
                              " restore_clients. If this is too permissive,"
                              " you may set a restore_client for individual"
                              " original clients in the individual"
                              " clientconfdir files. Note that restoring a"
                              " backup from a Windows computer onto a Linux"
                              " computer will currently leave the VSS headers"
                              " in place at the beginning of each file. This"
                              " will be addressed in a future version of"
                              " burp."),
        u'resume_partial': __(u"Turn this on to enable 'resume partial' code."
                              " Requires 'working_dir_recovery_method=resume'."
                              " When resuming an interrupted transfer of a"
                              " single file, it attempts to use previously"
                              " transferred blocks of that file in order to be"
                              " more efficient. However, situations have been"
                              " reported where the file on the server side"
                              " just gets bigger forever, so this feature now"
                              " defaults to being turned off."),
        u'scan_problem_raises_error': __(u"When enabled, this causes problems"
                                         " in the phase1 scan (such as an"
                                         " 'include' being missing) to be"
                                         " treated as fatal errors. The"
                                         " default is off."),
        u'server_script_arg': __(u"Goes with server_script and overrides"
                                 " server_script_pre_arg and"
                                 " server_script_post_arg."),
        u'server_script_notify': __(u"Turn on to send a notification emails"
                                    " when the server pre and post scripts"
                                    " return non-zero. The output of the"
                                    " script will be included it the email."
                                    " The default is off. Requires the"
                                    " notify_failure options to be set."),
        u'server_script_post_arg': __(u"A user-definable argument to the"
                                      " server post script. You can have many"
                                      " of these."),
        u'server_script_post_notify': __(u"Turn on to send a notification"
                                         " email when the server post script"
                                         " returns non-zero. The output of the"
                                         " script will be included in the"
                                         " email. The default is off. Requires"
                                         " the notify_failure options to be"
                                         " set."),
        u'server_script_post_run_on_fail': __(u"If this is set to 1,"
                                              " server_script_post will always"
                                              " be run. The default is 0,"
                                              " which means that if the task"
                                              " asked for by the client fails,"
                                              " server_script_post will not be"
                                              " run."),
        u'server_script_post': __(u"Path to a script to run on the server"
                                  " before the client disconnects. The"
                                  " arguments to it are 'post', '(client"
                                  " command)', '(client name), '(0 or 1 for"
                                  " success or failure)', '(timer script exit"
                                  " code)', and"
                                  " then arguments defined by"
                                  " server_script_post_arg. This command and"
                                  " related options can be overriddden by the"
                                  " client configuration files in"
                                  " clientconfdir on the server."),
        u'server_script_pre_arg': __(u"A user-definable argument to the server"
                                     " pre script. You can have many of"
                                     " these."),
        u'server_script_pre_notify': __(u"Turn on to send a notification email"
                                        " when the server pre script returns"
                                        " non-zero. The output of the script"
                                        " will be included in the email. The"
                                        " default is off. Most people will not"
                                        " want this turned on because clients"
                                        " usually contact the server at 20"
                                        " minute intervals and this could"
                                        " cause a lot of emails to be"
                                        " generated. Requires the"
                                        " notify_failure options to be set."),
        u'server_script_pre': __(u"Path to a script to run on the server after"
                                 " each successfully authenticated connection"
                                 " but before any work is carried out. The"
                                 " arguments to it are 'pre', '(client"
                                 " command)', '(client name)', '(0 or 1 for"
                                 " success or failure)', '(timer script exit"
                                 " code)', and"
                                 " then arguments defined by"
                                 " server_script_pre_arg. If the script"
                                 " returns non-zero, the task asked for by the"
                                 " client will not be run. This command and"
                                 " related options can be overriddden by the"
                                 " client configuration files in clientconfdir"
                                 " on the server."),
        u'server_script': __(u"You can use this to save space in your config"
                             " file when you want to run the same server"
                             " script twice. It overrides server_script_pre"
                             " and server_script_post. This command and"
                             " related options can be overriddden by the"
                             " client configuration files in clientconfdir on"
                             " the server."),
        u'soft_quota': __(u"A warning will be issued when the estimated size"
                          " of all files is greater than the specified size"
                          " and smaller than hard_quota. Example: 'soft_quota"
                          " = 95Gb'. Set to 0 (the default) to have no"
                          " warning."),
        u'split_vss': __(u"When backing up Windows computers with burp"
                         " protocol 1, this option allows you to save the VSS"
                         " header data separate from the file data. The"
                         " default is off, which means that the VSS header"
                         " data is saved prepended to the file data."),
        u'ssl_cert_ca': __(u"The path to the SSL CA certificate. This file"
                           " will probably be the same on both the server and"
                           " the client. The file should contain just the"
                           " certificate in PEM format. For more information"
                           " on this, and the other ssl_* options, please see"
                           " <a href='http://burp.grke.org/docs/burp_ca.html'>"
                           " docs/burp_ca.txt</a>."),
        u'ssl_cert_password': __(u"Synonym for ssl_key_password."),
        u'ssl_cert': __(u"The path to the server SSL certificate. It works for"
                        " me when the file contains the concatenation of the"
                        " certificate and private key in PEM format."),
        u'ssl_ciphers': __(u"Allowed SSL ciphers. See openssl ciphers for"
                           " details."),
        u'ssl_compression': __(u"Choose the level of zlib compression over"
                               " SSL. Setting 0 or zlib0 turnsSSL compression"
                               " off. Setting non-zero gives zlib5 compression"
                               " (it is not currently possible for openssl to"
                               " set any other level). The default is 5."
                               " 'gzip' is a synonym of 'zlib'.is a synonym of"
                               " 'zlib'."),
        u'ssl_dhfile': __(u"Path to Diffie-Hellman parameter file. To generate"
                          " one with openssl, use a command like this: openssl"
                          " dhparam -out dhfile.pem -5 1024"),
        u'ssl_key_password': __(u"The SSL key password."),
        u'ssl_key': __(u"The path to the server SSL private key in PEM"
                       " format."),
        u'status_address': __(u"Defines the main TCP address that the server"
                              " listens on for status requests. The default is"
                              " either '::1' or '127.0.0.1', dependent upon"
                              " compile time options."),
        u'status_port': __(u"Defines the TCP port that the server listens on"
                           " for status requests."),
        u'stdout': __(u"Log to stdout. Defaults to on."),
        u'strip_vss': __(u"When backing up Windows computers with burp"
                         " protocol 1, this option allows you to prevent the"
                         " VSS header data being backed up. The default is"
                         " off. To restore a backup that has no VSS"
                         " information on Windows, you need to give the client"
                         " the '-x' command line option."),
        u'syslog': __(u"Log to syslog. Defaults to off."),
        u'timer_arg': __(u"A user-definable argument to the timer script."
                         "You can have many of these. The timer_arg options"
                         " can be overridden by the client configuration files"
                         " in clientconfdir on the server."),
        u'timer_script': __(u"Path to the script to run when a client connects"
                            " with the timed backup option. If the script"
                            " exits with code 0, a backup will run. The first"
                            " two arguments are the client name and the path"
                            " to the 'current' storage directory. The next"
                            " three arguments are reserved, and user arguments"
                            " are appended after that. An example timer script"
                            " is provided. The timer_script option can be"
                            " overridden by the client configuration files in"
                            " clientconfdir on the server."),
        u'timestamp_format': __(u"This allows you to tweak the format of the"
                                " timestamps of individual backups. See 'man"
                                " strftime' to see available substitutions."
                                " If this option is unset, burp uses"
                                " \"%Y-%m-%d %H:%M:%S\"."),
        u'umask': __(u"Set the file creation umask. Default is 0022."),
        u'user': __(u"Run as a particular user. This can be overridden by the"
                    " client configuration files in clientconfdir on the"
                    " server."),
        u'version_warn': __(u"When this is on, which is the default, a warning"
                            " will be issued when the client version does not"
                            " match the server version. This option can be"
                            " overridden by the client configuration files in"
                            " clientconfdir on the server."),
        u'vss_drives': __(u"When backing up Windows computers, this option"
                          " allows you to specify which drives have VSS"
                          " snapshots taken of them. If you omit this option,"
                          " burp will automatically decide based on the"
                          " 'include' options. If you want no drives to have"
                          " snapshots taken of them, you can specify '0'."),
        u'working_dir_recovery_method': __(u"This option tells the server what"
                                           " to do when it finds the working"
                                           " directory of an interrupted"
                                           " backup (perhaps somebody pulled"
                                           " the plug on the server, or"
                                           " something). This can be"
                                           " overridden by the client"
                                           " configurations files in"
                                           " clientconfdir on the server."
                                           " Options are... <ul><li>delete:"
                                           " Just delete the old working"
                                           " directory.</li><li>use: Convert"
                                           " the working directory into a"
                                           " complete backup.</li><li>resume:"
                                           " Simply continue the previous"
                                           " backup from the point at which it"
                                           " left off, at file granularity."
                                           " NOTE: If the client has changed"
                                           " its include/exclude configuration"
                                           " since the backup was interrupted,"
                                           " the recovery method will"
                                           " automatically switch to 'use'."
                                           " </li></ul>"),
    }



# -*- coding: utf8 -*-
"""
.. module:: burpui._rtfd
    :platform: Unix
    :synopsis: Burp-UI wrapper documentation module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""

# THIS IS ONLY USED FOR DOC GENERATION

from . import create_app

# This is a lie we are not really unittesting, but we want to avoid the v2
# errors
app = create_app(conf='/dev/null', gunicorn=False, unittest=True)



# -*- coding: utf8 -*-
"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.cli
    :platform: Unix
    :synopsis: Burp-UI CLI module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import sys
import click
import subprocess

from .app import create_app

DEBUG = os.environ.get('BUI_DEBUG') or os.environ.get('FLASK_DEBUG') or False
if DEBUG and DEBUG.lower() in ['true', 'yes', '1']:
    DEBUG = True

VERBOSE = os.environ.get('BUI_VERBOSE') or 0
if VERBOSE:
    try:
        VERBOSE = int(VERBOSE)
    except ValueError:
        VERBOSE = 0

# UNITTEST is used to skip the burp-2 requirements for modes != server
UNITTEST = os.environ.get('BUI_MODE') not in ['server', 'manage', 'celery', 'legacy']
CLI = os.environ.get('BUI_MODE') not in ['server', 'legacy']

app = create_app(
    conf=os.environ.get('BUI_CONFIG'),
    verbose=VERBOSE,
    logfile=os.environ.get('BUI_LOGFILE'),
    debug=DEBUG,
    gunicorn=False,
    unittest=UNITTEST,
    cli=CLI
)

try:
    from .app import create_db
    from .ext.sql import db
    from flask_migrate import Migrate

    app.config['WITH_SQL'] = True
    create_db(app, True)

    mig_dir = os.getenv('BUI_MIGRATIONS')
    if mig_dir:
        migrate = Migrate(app, db, mig_dir)
    else:
        migrate = Migrate(app, db)
except ImportError:
    pass


@app.cli.command()
def legacy():
    """Legacy server for backward compatibility"""
    app.manual_run()


@app.cli.command()
@click.option('-b', '--backend', default='BASIC',
              help='User Backend (default is BASIC).')
@click.option('-p', '--password', help='Password to assign to user.',
              default=None)
@click.option('-a', '--ask', default=False, is_flag=True,
              help='If no password is provided and this flag is enabled, '
                   'you\'ll be prompted for one, else a random one will be '
                   'generated.')
@click.option('-v', '--verbose', default=False, is_flag=True,
              help='Add extra debug messages.')
@click.argument('name')
def create_user(backend, password, ask, verbose, name):
    """Create a new user."""
    app.load_modules(False)

    click.echo(click.style('[*] Adding \'{}\' user...'.format(name), fg='blue'))
    try:
        handler = getattr(app, 'uhandler')
    except AttributeError:
        handler = None

    if not handler or len(handler.backends) == 0 or \
            backend not in handler.backends:
        click.echo(click.style('[!] No authentication backend found', fg='red'))
        sys.exit(1)

    back = handler.backends[backend]

    if back.add_user is False:
        click.echo(click.style("[!] The '{}' backend does not support user "
                               "creation".format(backend), fg='red'))
        sys.exit(2)

    if not password:
        if ask:
            import getpass
            password = getpass.getpass()
            confirm = getpass.getpass('Confirm: ')
            if password != confirm:
                click.echo(click.style("[!] Passwords missmatch", fg='red'))
                sys.exit(3)
        else:
            import random

            alphabet = "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLM" \
                       "NOPQRSTUVWXYZ"
            pw_length = 8
            mypw = ""

            for i in range(pw_length):
                next_index = random.randrange(len(alphabet))
                mypw += alphabet[next_index]
            password = mypw
            click.echo(
                click.style(
                    '[+] Generated password: {}'.format(password),
                    fg='blue'
                )
            )

    success, message, _ = back.add_user(name, password)
    click.echo(click.style(
        '[+] Success: {}{}'.format(
            success, ' -> {}'.format(message) if verbose and message else ''
        ),
        fg='green' if success else 'red')
    )


@app.cli.command()
@click.argument('language')
def init_translation(language):
    """Initialize a new translation for the given language."""
    os.system('pybabel extract -F babel.cfg -k __ -k lazy_gettext -o messages.pot burpui')
    os.system('pybabel init -i messages.pot -d burpui/translations -l {}'.format(language))
    os.unlink('messages.pot')


@app.cli.command()
def update_translation():
    """Update translation files."""
    os.system('pybabel extract -F babel.cfg -k __ -k lazy_gettext -o messages.pot burpui')
    os.system('pybabel update -i messages.pot -d burpui/translations')
    os.unlink('messages.pot')


@app.cli.command()
def compile_translation():
    """Compile translations."""
    os.system('pybabel compile -f -d burpui/translations')


@app.cli.command()
@click.option('-b', '--burp-conf-cli', 'bconfcli', default=None,
              help='Burp client configuration file')
@click.option('-s', '--burp-conf-serv', 'bconfsrv', default=None,
              help='Burp server configuration file')
@click.option('-c', '--client', default='bui',
              help='Name of the burp client that will be used by Burp-UI '
                   '(defaults to "bui")')
@click.option('-h', '--host', default='::1',
              help='Address of the status server (defaults to "::1")')
@click.option('-r', '--redis', default=None,
              help='Redis URL to connect to')
@click.option('-d', '--database', default=None,
              help='Database to connect to for persistent storage')
@click.option('-n', '--dry', is_flag=True,
              help='Dry mode. Do not edit the files but display changes')
def setup_burp(bconfcli, bconfsrv, client, host, redis, database, dry):
    """Setup burp client for burp-ui."""
    if app.vers != 2:
        click.echo(
            click.style(
                'Sorry, you can only setup the Burp 2 client',
                fg='red'
            ),
            err=True
        )
        sys.exit(1)

    if not app.standalone:
        click.echo(
            click.style(
                'Sorry, only the standalone mode is supported',
                fg='red'
            ),
            err=True
        )
        sys.exit(1)

    app.load_modules(False)

    from .misc.parser.utils import Config
    from .app import get_redis_server
    import difflib
    import tempfile

    orig = source = None
    conf_orig = []
    if dry:
        try:
            with open(app.conf.options.filename) as fil:
                conf_orig = fil.readlines()
        except:
            pass

        orig = source = app.conf.options.filename
        (_, temp) = tempfile.mkstemp()
        app.conf.options.filename = temp

    if not app.conf.lookup_section('Burp2', source):
        app.conf._refresh(True)
    if (database or redis) and not app.conf.lookup_section('Production', source):
        app.conf._refresh(True)

    def _edit_conf(key, val, attr):
        if val and (((key not in app.conf.options['Burp2']) or
                    (key in app.conf.options['Burp2'] and
                    val != app.conf.options['Burp2'][key])) and
                    getattr(app.client, attr) != val):
            app.conf.options['Burp2'][key] = val
            app.conf.options.write()
            app.conf._refresh(True)

    def _color_diff(line):
        if line.startswith('+'):
            return click.style(line, fg='green')
        elif line.startswith('-'):
            return click.style(line, fg='red')
        elif line.startswith('^'):
            return click.style(line, fg='blue')
        return line

    _edit_conf('bconfcli', bconfcli, 'burpconfcli')
    _edit_conf('bconfsrv', bconfsrv, 'burpconfsrv')

    if redis:
        try:
            # detect missing modules
            import redis as redis_client  # noqa
            import celery  # noqa
            if ('redis' not in app.conf.options['Production'] or
                'redis' in app.conf.options['Production'] and
                app.conf.options['Production']['redis'] != redis) and \
                    app.redis != redis:
                app.conf.options['Production']['redis'] = redis

            rhost, rport, _ = get_redis_server(app)
            DEVNULL = open(os.devnull, 'wb')
            ret = subprocess.call(['/bin/nc', '-z', '-w5', str(rhost), str(rport)], stdout=DEVNULL, stderr=subprocess.STDOUT)

            if ret == 0:
                app.conf.options['Production']['celery'] = 'true'

                app.conf.options['Production']['storage'] = 'redis'

                app.conf.options['Production']['cache'] = 'redis'
            else:
                click.echo(
                    click.style(
                        'Unable to contact the redis server, disabling it',
                        fg='yellow'
                    )
                )
                app.conf.options['Production']['storage'] = 'default'
                app.conf.options['Production']['cache'] = 'default'
                if app.use_celery:
                    app.conf.options['Production']['celery'] = 'false'

            app.conf.options.write()
            app.conf._refresh(True)
        except ImportError:
            click.echo(
                click.style(
                    'Unable to activate redis & celery. Did you ran the '
                    '\'pip install burp-ui[celery]\' and '
                    '\'pip install burp-ui[gunicorn-extra]\' commands first?',
                    fg='yellow'
                )
            )

    if database:
        try:
            from .ext.sql import db  # noqa
            if ('database' not in app.conf.options['Production'] or
                'database' in app.conf.options['Production'] and
                app.conf.options['Production']['database'] != database) and \
                    app.database != database:
                app.conf.options['Production']['database'] = database
                app.conf.options.write()
                app.conf._refresh(True)
        except ImportError:
            click.echo(
                click.style(
                    'It looks like some dependencies are missing. Did you ran '
                    'the \'pip install burp-ui[sql]\' command first?',
                    fg='yellow'
                )
            )

    if dry:
        temp = app.conf.options.filename
        app.conf.options.filename = orig
        after = []
        try:
            if not os.path.exists(temp) or os.path.getsize(temp) == 0:
                after = conf_orig
            else:
                with open(temp) as fil:
                    after = fil.readlines()
                os.unlink(temp)
        except:
            pass
        diff = difflib.unified_diff(conf_orig, after, fromfile=orig, tofile='{}.new'.format(orig))
        out = ''
        for line in diff:
            out += _color_diff(line)
        if out:
            click.echo_via_pager(out)

    bconfcli = bconfcli or app.conf.options['Burp2'].get('bconfcli') or \
        getattr(app.client, 'burpconfcli')
    bconfsrv = bconfsrv or app.conf.options['Burp2'].get('bconfsrv') or \
        getattr(app.client, 'burpconfsrv')
    dest_bconfcli = bconfcli

    if not os.path.exists(bconfcli):
        clitpl = """
mode = client
port = 4971
status_port = 4972
server = ::1
password = abcdefgh
cname = {0}
protocol = 1
pidfile = /tmp/burp.client.pid
syslog = 0
stdout = 1
progress_counter = 1
network_timeout = 72000
server_can_restore = 0
cross_all_filesystems=0
ca_burp_ca = /usr/sbin/burp_ca
ca_csr_dir = /etc/burp/CA-client
ssl_cert_ca = /etc/burp/ssl_cert_ca-client-{0}.pem
ssl_cert = /etc/burp/ssl_cert-bui-client.pem
ssl_key = /etc/burp/ssl_cert-bui-client.key
ssl_key_password = password
ssl_peer_cn = burpserver
include = /home
exclude_fs = sysfs
exclude_fs = tmpfs
nobackup = .nobackup
exclude_comp=bz2
exclude_comp=gz
""".format(client)

        if dry:
            (_, dest_bconfcli) = tempfile.mkstemp()
        with open(dest_bconfcli, 'w') as confcli:
            confcli.write(clitpl)

    parser = app.client.get_parser()

    confcli = Config(dest_bconfcli, parser, 'srv')
    confcli.set_default(dest_bconfcli)
    confcli.parse()

    if confcli.get('cname') != client:
        confcli['cname'] = client
    if confcli.get('server') != host:
        confcli['server'] = host

    if confcli.dirty:
        if dry:
            (_, dstfile) = tempfile.mkstemp()
        else:
            dstfile = bconfcli

        confcli.store(dest=dstfile, insecure=True)
        if dry:
            before = []
            after = []
            try:
                with open(bconfcli) as fil:
                    before = fil.readlines()
            except:
                pass
            try:
                with open(dstfile) as fil:
                    after = fil.readlines()
                os.unlink(dstfile)
            except:
                pass

            if dest_bconfcli != bconfcli:
                # the file did not exist
                os.unlink(dest_bconfcli)
                before = []

            diff = difflib.unified_diff(before, after, fromfile=bconfcli, tofile='{}.new'.format(bconfcli))
            out = ''
            for line in diff:
                out += _color_diff(line)
            if out:
                click.echo_via_pager(out)

    if not os.path.exists(bconfsrv):
        click.echo(
            click.style(
                'Unable to locate burp-server configuration, aborting!',
                fg='red'
            ),
            err=True
        )
        sys.exit(1)

    confsrv = Config(bconfsrv, parser, 'srv')
    confsrv.set_default(bconfsrv)
    confsrv.parse()

    if host not in ['::1', '127.0.0.1']:
        bind = confsrv.get('status_address')
        if (bind and bind not in [host, '::', '0.0.0.0']) or not bind:
            click.echo(
                click.style(
                    'It looks like your burp server is not exposing it\'s '
                    'status port in a way that is reachable by Burp-UI!',
                    fg='yellow'
                )
            )
            click.echo(
                click.style(
                    'You may want to set the \'status_address\' setting with '
                    'either \'{}\', \'::\' or \'0.0.0.0\' in the {} file '
                    'in order to make Burp-UI work'.format(host, bconfsrv),
                    fg='blue'
                )
            )

    if 'restore_client' not in confsrv:
        confsrv['restore_client'] = client
    else:
        restore = confsrv.getlist('restore_client')
        if client not in restore:
            confsrv['restore_client'].append(client)

    confsrv['monitor_browse_cache'] = True

    ca_client_dir = confsrv.get('ca_csr_dir')
    if ca_client_dir and not os.path.exists(ca_client_dir):
        try:
            os.makedirs(ca_client_dir)
        except IOError as exp:
            click.echo(
                click.style(
                    'Unable to create "{}" dir: {}'.format(ca_client_dir, exp),
                    fg='yellow'
                ),
                err=True
            )

    if confsrv.dirty:
        if dry:
            (_, dstfile) = tempfile.mkstemp()
        else:
            dstfile = bconfsrv

        confsrv.store(dest=dstfile, insecure=True)
        if dry:
            before = []
            after = []
            try:
                with open(bconfsrv) as fil:
                    before = fil.readlines()
            except:
                pass
            try:
                with open(dstfile) as fil:
                    after = fil.readlines()
                os.unlink(dstfile)
            except:
                pass
            diff = difflib.unified_diff(before, after, fromfile=bconfsrv, tofile='{}.new'.format(bconfsrv))
            out = ''
            for line in diff:
                out += _color_diff(line)
            if out:
                click.echo_via_pager(out)

    if confsrv.get('clientconfdir'):
        bconfagent = os.path.join(confsrv.get('clientconfdir'), client)
    else:
        click.echo(
            click.style(
                'Unable to find "clientconfdir" option, you will have to '
                'setup the agent by your own',
                fg='yellow'
            )
        )
        bconfagent = os.devnull

    if not os.path.exists(bconfagent):

        agenttpl = """
password = abcdefgh
"""

        if not dry:
            with open(bconfagent, 'w') as confagent:
                confagent.write(agenttpl)
        else:
            before = []
            after = ['{}\n'.format(x) for x in agenttpl.splitlines()]
            diff = difflib.unified_diff(before, after, fromfile='None', tofile=bconfagent)
            out = ''
            for line in diff:
                out += _color_diff(line)
            if out:
                click.echo_via_pager(out)

    else:
        confagent = Config(bconfagent, parser, 'cli')
        confagent.set_default(bconfagent)
        confagent.parse()

        if confagent.get('password') != confcli.get('password'):
            click.echo(
                click.style(
                    'It looks like the passwords in the {} and the {} files '
                    'missmatch. Burp-UI will not work properly untill you fix '
                    'this'.format(bconfcli, bconfagent),
                    fg='yellow'
                )
            )
